from graph import Graph
from itertools import count


class SchedulingError(ValueError):
    pass


class NullTask(object):
    scheduled_finish = 0


class Task(object):
    _ids = count()

    def __init__(self, requires, client=None, supplier=None, earliest_start=None, latest_finish=None,
                 latest_start=None, earliest_finish=None, duration=None, cost=None):
        """
        :param requires: multiple types:
            dicts are taken as given.
            sets, lists, tuples are transformed to dicts with key: count(key)
            strings are handled as {str: 1}
        :param client: the resource requiring the task to be completed.
        :param earliest_start: sortable value
        :param latest_finish: sortable value
        """
        if isinstance(requires, (set, tuple, list)):
            self.requires = {k: requires.count(k) for k in set(requires)}
        elif isinstance(requires, str):
            self.requires = {requires: 1}
        elif isinstance(requires, dict):
            self.requires = requires
        else:
            raise TypeError(f"requires expected dict, tuple, list, set or str, but got {type(requires)}")

        if client is None or isinstance(client, int):
            self.client = client
        else:
            raise TypeError(f"expected client to be int, not {type(client)}")

        if supplier is None or isinstance(supplier, int):
            self.supplier = supplier
        else:
            raise TypeError(f"expected supplier to be int, not {type(client)}")

        self.earliest_start = earliest_start
        self.earliest_finish = earliest_finish
        self.latest_start = latest_start
        self.latest_finish = latest_finish
        self.duration = duration
        self.cost = cost

        self.scheduled_start = None
        self.scheduled_finish = None

        self.id = next(self._ids)

    def __eq__(self, other):
        """ returns True if process.output.keys() == task.requires.keys()"""
        if isinstance(other, Process):
            return self.requires.keys() == other.outputs.keys()
        elif isinstance(other, Task):
            return self.requires.keys() == other.requires.keys()
        else:
            raise NotImplementedError  # ... if we haven't returned above ...

    def __bool__(self):
        return self.scheduled_start is not None and self.scheduled_finish is not None

    def __str__(self):
        return f"{self.__class__.__name__}({self.id}) {self.requires}"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return self.id

    def copy(self):
        return Task(requires=self.requires,
                    client=self.client,
                    earliest_start=self.earliest_start,
                    latest_finish=self.latest_finish,
                    latest_start=self.latest_start,
                    earliest_finish=self.earliest_finish,
                    duration=self.duration,
                    cost=self.cost)


class Process(object):
    def __init__(self, inputs=None, outputs=None,
                 setup_time=0, run_time=0, shutdown_time=0, change_over_time=0, cost=0):
        """
        :param inputs: multiple types:
            dicts are taken as given.
            sets, lists, tuples are transformed to dicts with key: count(key)
            strings are handled as {str: 1}
        :param outputs: multiple types:
            dicts are taken as given.
            sets, lists, tuples are transformed to dicts with key: count(key)
            strings are handled as {str: 1}
        :param setup_time: any sortable value
        :param run_time: any sortable value
        :param shutdown_time: any sortable value
        :param change_over_time: any sortable value
        :param cost: any sortable value
        """
        if inputs is None:
            self.inputs = {}
        elif isinstance(inputs, (set, tuple, list)):
            self.inputs = {k: inputs.count(k) for k in set(inputs)}
        elif isinstance(inputs, str):
            self.inputs = {inputs: 1}
        elif isinstance(inputs, dict):
            self.inputs = inputs
        else:
            raise TypeError(f"inputs expected set, tuple, list or dict, but got {type(inputs)}")

        if outputs is None:
            self.outputs = {}
        elif isinstance(outputs, (set, tuple, list)):
            self.outputs = {k: outputs.count(k) for k in set(outputs)}
        elif isinstance(outputs, str):
            self.outputs = {outputs: 1}
        elif isinstance(outputs, dict):
            self.outputs = outputs
        else:
            raise ValueError(f"outputs expected set, tuple, list or dict, but got {type(outputs)}")

        for i in [setup_time, run_time, shutdown_time, change_over_time, cost]:
            if not isinstance(i, (int, float)):
                raise TypeError(f"{i.__name__} is {type(i)}. Expected float or int")

        # The values below are set by the resource.
        self.setup_time = setup_time
        self.run_time = run_time
        self.shutdown_time = shutdown_time
        self.change_over_time = change_over_time
        self.cost = cost

    def __str__(self):
        return f"{self.__class__.__name__}: {self.outputs} <- {self.inputs}"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        """ return True if process.output.keys() == task.requires.keys()
        Note:
        As one 'production' could supply multiple tasks, the challenge is to
        match tasks with productions, but should probably be performed in
        another function.
        """
        if isinstance(other, Task):
            return self.outputs.keys() == other.requires.keys()

        raise NotImplementedError  # ... if we haven't returned above ...


class Resource(object):
    """
    A Resource can perform a single process at any time.
    """
    _ids = count()

    def __init__(self):
        self.id = next(self._ids)
        self.new_tasks = []  # new tasks
        self.tasks = []  # known tasks
        self.supply = {}  # task = [list of tasks given to suppliers]
        self.processes = []
        self._rdn = None

    def __str__(self):
        return f"{self.__class__.__name__}({self.id})"

    def __hash__(self):
        return self.id

    @property
    def rdn(self):
        return self._rdn

    @rdn.setter
    def rdn(self, value):
        if not isinstance(value, ResourceDemandNetwork):
            err = f"{self.__class__.__name__}.rdn can only be set by the {ResourceDemandNetwork.__name__}"
            raise ValueError(err)
        if self._rdn is not None:
            raise ValueError(f"{self}.rdn has already been set.")
        self._rdn = value

    def add_task(self, task):
        assert isinstance(task, Task)
        if self._rdn is None:
            raise ValueError(f"{self} has not been added to the RDN. Do this first.")
        if not any(p == task for p in self.processes):
            raise ValueError(f"{self} has no process supporting {task}")
        self.new_tasks.append(task)
        self._rdn.notify(r_id=self.id)

    def remove_task(self, task):
        assert isinstance(task, Task)
        if task in self.tasks:
            self.tasks.remove(task)
        if task in self.new_tasks:
            self.new_tasks.remove(task)

    def add_process(self, proc):
        assert isinstance(proc, Process)
        self.processes.append(proc)

    def supplies(self, task):
        for p in self.processes:
            if p == task:
                return p.inputs
        raise ValueError(f"{self} does not support {task}")

    def get_process(self, task):
        for p in self.processes:
            if p == task:
                return p
        return None

    def can_support(self, task):
        """ returns bool; True if there's a process that can support the task """
        if any(p == task for p in self.processes):
            return True
        return False

    def notify(self):
        """ api for other resources to ping, so that this resource is updated.

        when customers add tasks to resources, they use resource.add_task(Task).
        This method puts the id of the resource onto the RDNs message queue, so that
        the RDN will identify the resource and run its .schedule() method.

        The opposite direction is more indirect, as a resource may finish a schedule,
        but has no way of notifying the customer. To overcome this, the resource can
        identify the customer from the Task.client, and `notify` the client using this
        method.
        """
        if self._rdn is None:
            raise ValueError(f"{self} has not been added to the RDN. Do this first.")
        self._rdn.notify(r_id=self.id)

    def suppliers(self, task):
        """ returns list of suppliers who can support the task. """
        assert isinstance(task, Task)
        suppliers = []
        for nid in self._rdn.network.nodes(from_node=self.id):
            resource = self._rdn.network.node(nid)
            assert isinstance(resource, Resource)
            if resource.can_support(task):
                suppliers.append(resource)
        return suppliers

    def schedule(self):
        """ calculates the schedule. Stores the schedule in self.tasks

        1. discover tasks:
            tasks are in self.mq
            for each new task:
                find supply requirements (proc.output requires proc.input)
                find suppliers
                for each supplier:
                    create a new (supply) task as: Task(requires=supply, required_by=self.id)
                    add task to supplier (which automatically notifies)
                    notify via supplier.notify()

        2. determine better schedule (if all supply tasks are scheduled for delivery):
        sort tasks by arrival (as we hold no inventory)

        3. detect idle-time, lateness, ...
        remove any duplicate tasks (and ping suppliers to update delivery times).

        notify client of change in delivery times.

        """
        if not isinstance(self._rdn, ResourceDemandNetwork):
            raise AttributeError(f"Use ResourceDemandNetwork.add_resource({self}) first.")

        # 1. discover tasks
        while self.new_tasks:
            task = self.new_tasks.pop(0)
            assert isinstance(task, Task), type(task)
            self.tasks.append(task)
            process = self.get_process(task)
            assert isinstance(process, Process)
            if not process.inputs:
                continue  # nothing to do. supplies are not required.

            self.supply[task] = []  # create empty list for adding tasks given to suppliers.
            # below:
            # if there are more customers to a supplier, it'll be ambiguous, who's ordering what.
            # Therefore we set: Task.client = self.id.
            supply_task = Task(requires=process.inputs, client=self.id)

            for resource in self.suppliers(supply_task):
                new_task = Task(requires=process.inputs, client=self.id, supplier=resource.id)
                self.supply[task].append(new_task)  # own reference to task.
                resource.add_task(new_task)  # adding task suppliers inbox.
            return  # as this resource created tasks, it will have to wait for reply from suppliers

        for supply_task, supplier_list in self.supply.items():
            if any(not t for t in supplier_list):
                return  # resource is waiting for information from suppliers.

        # as self.supply contains all scheduled supply tasks, it is possible to iterate through own
        # tasks and determine the best supply option without having to find the supplier.
        # -- some kind of match(self.tasks, self.supply)

        # PS> remember to cancel duplicate tasks for stuff that isn't needed.

        # 2. check / determine schedule.

        # to minimise total idle time, sort tasks by (runtime, name)
        assert all(isinstance(t, Task) for t in self.tasks)  # this is disabled with python -OO
        tasks = [(t.runtime, t.name, t) for t in self.tasks]
        # `runtime` to minimise initial waiting time, and,
        # `name` to assure that task of same name come out together so c/o time can be exploited.
        tasks.sort()  # ascending!
        self.tasks = [t for r, n, t in tasks]

        # for task in tasks: update start and finish time.
        previous_task = NullTask
        for ix, task in enumerate(self.tasks):
            process = self.get_process(task)
            assert isinstance(process, Process)
            assert isinstance(task, Task)

            if previous_task == task:  # handling change_over time instead of shutdown time if the tasks are the same.
                previous_task.scheduled_finish += (process.change_over_time - process.shutdown_time)

            if process.inputs:  # pick best matching supply time.
                supply_times = [(t.scheduled_finish, t) for t in self.supply[task]]
                supply_times.sort()
                supply_times = [t for _, t in supply_times]

                fas = supply_times[0]  # pick first available supply
                for t in supply_times[1:]:  # cancel all other tasks
                    resource = self._rdn.network.node(node_id=t.supplier)
                    assert isinstance(resource, Resource)
                    resource.remove_task(t)
                    resource.notify()

                start_time = max(previous_task.scheduled_finish, fas.scheduled_finish)
            else:
                start_time = previous_task.scheduled_finish

            task.scheduled_start = start_time

            if previous_task == task:
                steps = [task.scheduled_start, process.run_time, process.shutdown_time]
            else:
                steps = [task.scheduled_start, process.setup_time, process.run_time, process.shutdown_time]
            task.scheduled_finish = sum(steps)

            # get ready to loop.
            previous_task = task

        # at this point, the tasks are all scheduled with start and finish times.

        # check if there's any idle time.
        # notify customer.

        print(f"{self} done")




class ResourceDemandNetwork(object):
    """ Resource-Demand Network
    Resources (machines, people, inventory) are connected in a network, where
    the edges permit transformation of tasks.

    Tasks must be given directly to Resources
    """
    def __init__(self):
        self.network = Graph()
        self.task_queue = {}  # only keys are used as this behaves like an ordered set

    def add_resource(self, resource):
        """ adds resource to RDN. """
        assert isinstance(resource, Resource)
        if resource.id not in self.network.nodes():
            self.network.add_node(node_id=resource.id, obj=resource)
            resource.rdn = self
        # else: ... just return.

    def add_edge(self, client, resource):
        """ A directed edge"""
        assert isinstance(client, Resource)
        assert isinstance(resource, Resource)
        self.network.add_node(node_id=client.id, obj=client)
        self.network.add_node(node_id=resource.id, obj=resource)
        self.network.add_edge(client.id, resource.id, value=1)  # hop size = 1.
        if client.rdn is None:
            client.rdn = self
        if resource.rdn is None:
            resource.rdn = self

    def notify(self, r_id):
        """ method for assure that RDN will run resource.schedule() """
        if r_id not in self.task_queue:
            self.task_queue[r_id] = self.network.node(r_id)
        # else: ... just return.

    def schedule(self):
        """ schedules the tasks on the resources """
        # 1. discover new tasks and prepare the task_queue:
        resource_ids = self.network.nodes()
        if not resource_ids:
            raise SchedulingError("No resources to schedule.")

        for r_id in resource_ids:
            resource = self.network.node(r_id)
            if resource.mq:
                self.task_queue[r_id] = resource
        # task_queue is now prepared.

        # 2. initiate the scheduling.
        while self.task_queue:
            # first swop the pointer for the queue, so I can loop over it, without
            # changing the order of events.
            current_queue, self.task_queue = self.task_queue, {}

            for r_id, resource in current_queue.items():
                resource = self.network.node(node_id=r_id)
                assert isinstance(resource, Resource)
                resource.schedule()
                # self.task_queue is now populated by:
                # - any resource that places tasks on any downstream resource
                # - any resource that finishes scheduling and uses notify





from graph import Graph
from itertools import count


class FiniteStateMachine(object):
    def __init__(self):
        self.states = Graph()
        self.current_state = None
        self._action_id = count()
        self.actions = {}
        self._initial_state_was_set = False

    def set_initial_state(self, state):
        """  method for setting initial state of the FSM

        :param state: available in options
        :return: None
        """
        if self._initial_state_was_set:
            raise ValueError(f"initial state has already been set.")
        if state not in self.states.nodes():
            raise ValueError(f"{state} is not a state.")
        self.current_state = state
        self._initial_state_was_set = True

    def add_transition(self, state_1, action, state_2):
        """ Adds a state transition from state 1 to state 2 if action is performed.

        :param state_1: any hashable value.
        :param action: any hashable value.
        :param state_2: any hashable value.
        :return: None
        """
        action_id = next(self._action_id)
        self.states.add_edge(node1=state_1, node2=action_id)
        self.states.add_edge(node1=action_id, node2=state_2)
        self.actions[action_id] = action

    def options(self):
        """ returns list of options for the FSMs current state. """
        return [self.actions[i] for i in self.states.nodes(from_node=self.current_state)]

    def next(self, action):
        """ transitions the FSM from it's current state as a reaction to input `action`.

        :param action: any action available in fsm.options()
        :return: None
        """
        if self._initial_state_was_set is False:
            raise ValueError("initial state has not been set.")
        actions = self.states.nodes(from_node=self.current_state)
        if not actions:
            raise StopIteration("terminal node reached.")

        for action_id in actions:
            if self.actions[action_id] == action:
                self.current_state = self.states.nodes(from_node=action_id)[0]
                return
        raise ValueError(f"{self.current_state} does not permit {action}.")




from graph import Graph
import itertools

__description__ = """
Transshipment problems form a subgroup of transportation problems, where 
transshipment is allowed. In transshipment, transportation may or must go 
through intermediate nodes, possibly changing modes of transport.
Transshipment or Transhipment is the shipment of goods or containers to an 
intermediate destination, and then from there to yet another destination. One 
possible reason is to change the means of transport during the journey (for 
example from ship transport to road transport), known as transloading. Another 
reason is to combine small shipments into a large shipment (consolidation), 
dividing the large shipment at the other end (deconsolidation). Transshipment 
usually takes place in transport hubs. Much international transshipment also 
takes place in designated customs areas, thus avoiding the need for customs 
checks or duties, otherwise a major hindrance for efficient transport.
[1](https://en.wikipedia.org/wiki/Transshipment_problem)
"""


def clondike_transshipment_problem():
    """
    A deep gold mining operation is running at full speed.

    Problem:
    What schedule guarantees an optimal throughput given unpredictable output
    from the mines and unpredictable needs for mining equipment?

    Constraints:
    The gold mine has a surface depot which connects the 4 mining levels using
    an elevator running in the main vertical mine shaft.
    At each level there is a narrow gage electric train which can pull rail cars
    between the lift and side tracks where excavation occurs.
    Loading minerals and unloading equipment takes time, so the best way to
    avoid obstructing the busy lift is by moving the rail cars directly into
    the lift. However as mining equipment and valuable minerals are very heavy,
    the lift can only move one load at the time.
    Similarly the best way to avoid obstructing the railway at each level is by
    by moving the narrow gage rail cars onto side tracks that are available at
    each horizontal mine entry.
    Schematic of the mine:
        Mining shaft lift to surface depot.
        ^
        |
        +- level 1 <------------------->
        |              1  2  3  ... N  (horizontal mine shafts)
        |
        +- level 2 <------------------->
        |              1  2  3  ... N
        |
        +- level 3 <------------------->
        |              1  2  3  ... N
        |
        +- level 4 <------------------->
        |              1  2  3  ... N
        |
        +- level N
    In the intersection between the mining shaft lift and the electric
    locomotive there is limited space to perform any exchange, and because of
    constraints of the equipment, the order of delivery is strict:
       ^ towards surface.
       |
       |            (railway switch)
       |                |
    [ from lift ] ---->[S]<-->[ electric locomotive ]<---> into the mine.
    [ onto lift ] <-----|
       |
       |
       v into the mine.
    """
    paths = [
        ("Surface", "L-1", 1),
        ("L-1", "L-2", 1),
        ("L-2", "L-3", 1),
        ("L-3", "L-4", 1),
        ("L-1", "L-1-1", 1),
        ("L-2", "L-2-1", 1),
        ("L-3", "L-3-1", 1),
        ("L-4", "L-4-1", 1),
    ]

    for level in [1, 2, 3, 4]:  # adding stops for the narrow gage trains in the levels.
        paths.append(("L-{}".format(level), "L-{}-1".format(level), 1), )
        for dig in [1, 2, 3, 4, 5, 6]:
            paths.append(("L-{}-{}".format(level, dig), "L-{}-{}".format(level, dig + 1), 1))

    paths.extend([(n2, n1, d) for n1, n2, d in paths])  # adding the reverse path.
    g = Graph(from_list=paths)
    return g


class Train(object):
    def __init__(self, rail_network, start_location, access):
        """
        :param rail_network: the while rail network as a Graph.
        :param start_location: a node in the network.
        :param access: Set of nodes to which this train has access.
        """
        assert isinstance(rail_network, Graph)
        self._rail_network = rail_network
        assert start_location in self._rail_network.nodes()
        self._current_location = start_location
        assert isinstance(access, set)
        assert all([n in self._rail_network for n in access])
        self._access_nodes = access

        self._schedule = []

    def schedule(self, jobs=None):
        """
        Initialise the solution using shortest jobs first.
        Then improve the solution using combinatorics until improvement is zero
        param: jobs, (optional) list of jobs
        returns: list of jobs in scheduled order.
        """
        if jobs is None:
            return self._schedule
        assert isinstance(jobs, list)
        new_jobs = find(rail_network=self._rail_network, stops=self._access_nodes, jobs=jobs)
        self._schedule = schedule(graph=self._rail_network, start=self._current_location, jobs=new_jobs)
        return self._schedule


def schedule_rail_system(rail_network, trains, jobs):
    """
    Iterative over the trains until a feasible schedule is found.
    1. Each device loop through the jobs:
        select all relevant jobs.
        create itinerary items on the jobs.
        when all itineraries are complete, return the schedule.
    """
    assert isinstance(rail_network, Graph)
    assert isinstance(trains, list)
    assert all(isinstance(t, Train) for t in trains)
    assert isinstance(jobs, list)

    for train in trains:
        assert isinstance(train, Train)
        train.schedule(jobs)


def find(rail_network, stops, jobs):
    """
    Finds the route sections that the jobs need to travel from/to (stops)
    in the rail network.
    :param rail_network: class Graph
    :param stops: set of train stops
    :param jobs: list of jobs.
    :return: sub jobs.
    """
    sub_jobs = []
    for A, B in jobs:
        if A not in rail_network:
            raise ValueError("{} not in rail network".format(A))
        if B not in rail_network:
            raise ValueError("{} not in rail network".format(B))

        _, path = rail_network.shortest_path(A, B)
        part_route = [p for p in path if p in stops]

        if not rail_network.has_path(part_route):
            raise ValueError("Can't find path for {}".format(part_route))

        new_a, new_b = part_route[0], part_route[-1]
        sub_jobs.append((new_a, new_b))
    return sub_jobs


def schedule(graph, start, jobs):
    """
    The best possible path is a circuit.
    First we'll find all circuits and attempt to remove them from
    the equation.
    Once no more circuits can be found, the best solution is to look for
    alternative combinations that provide improvement.
    :return:
    """
    new_schedule = []
    jobs_to_plan = jobs[:]
    while jobs_to_plan:
        circuit_path = find_perfect_circuit(graph=graph, start=start, jobs=jobs_to_plan)
        if circuit_path:
            job_sequence = jobs_from_path(circuit_path)
        else:  # circuit not possible.
            shortest_path = []
            shortest_distance = float('inf')
            for perm in itertools.permutations(jobs_to_plan, len(jobs_to_plan)):
                path = path_from_schedule(jobs=perm, start=start)
                distance = graph.distance_from_path(path)
                if distance < shortest_distance:
                    shortest_distance = distance
                    shortest_path = path
            job_sequence = jobs_from_path(shortest_path)
        # remove planned jobs from options:
        for job in job_sequence:
            if job in jobs_to_plan:
                jobs_to_plan.remove(job)
                new_schedule.append(job)
    return new_schedule


def jobs_from_path(path):
    """ helper for finding jobs from path"""
    return [(path[i], path[i + 1]) for i in range(len(path) - 1)]


def path_from_schedule(jobs, start):
    """ The evaluation is based on building the travel path.
    For example in the network A,B,C with 4 trips as:
        1 (A,B), 2 (A,C), 3 (B,A), 4 (C,A)
    which have the travel path: [A,B,A,C,B,A,C,A]
    The shortest path for these jobs is: [A,C,A,B,A] which uses the order:
        2 (A,C), 4 (C,A), 1 (A,B), 3(B,A)
    """
    path = [start]
    for A, B in jobs:
        if A != path[-1]:
            path.append(A)
        path.append(B)
    return path


def find_perfect_circuit(graph, start, jobs):
    """ A perfect circuit is a path that starts and ends at the same place
    and where every movement includes a job.
    :param: start: starting location.
    :param: jobs: list of movements [(A1,B1), (A2,B2,) ....]
    :return path [A1,B1, ..., A1]
    """
    g = Graph()
    for A, B in jobs:
        try:
            g.edge(A, B)
        except KeyError:
            d, p = graph.shortest_path(A, B)
            g.add_edge(A, B, d)

    new_starts = [B for A, B in jobs if A == start]
    for A in new_starts:
        if A in g:
            p = g.breadth_first_search(A, start)  # path back to start.
            if p:
                return [start] + p
    return []




try:
    from matplotlib import pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D  # import required by matplotlib.
    visuals_enabled = True
except ImportError:
    visuals_enabled = False


def visualise(func):
    def wrapper(*args, **kwargs):
        if not visuals_enabled:
            raise ImportError("visualise is not available unless matplotlib is installed")
        return func(*args, **kwargs)
    return wrapper


@visualise
def plot_2d(graph, nodes=True, edges=True):
    """
    :param graph: instance of Graph with nodes as (x,y)
    :param nodes: bool: plots nodes
    :param edges: bool: plots edges
    :return: matlibplot.pyplot

    PRO-TIP: If your graph does not have nodes as (x,y) use random_xy_graph to
    create it with this recipe:

    Step 1: Get the imports.
    >>> from graph.random import random_xy_graph
    >>> from graph.hash import graph_hash

    Step 2: Determine the initialisation values.
    >>> node_count = len(graph.nodes())
    >>> seed = graph_hash(graph)
    >>> x_max = node_count * 8
    >>> y_max = node_count * 4
    >>> xygraph = random_xy_graph(node_count, x_max, y_max, edges=0, seed=seed)

    Step 3: create a mapping between the xy graph and the original graph.
    >>> mapping = {a:b for a,b in zip(xygraph.nodes(), graph.nodes())}

    Step 4: add the edges:
    >>> for edge in graph.edges():
    >>>     start, end, distance = edge
    >>>     xygraph.add_edge(mapping[start], mapping[end], distance)
    >>> plt = xygraph.plot_2d()
    >>> plt.show()

    """
    assert isinstance(nodes, bool)
    assert isinstance(edges, bool)

    for node in graph.nodes():
        if not isinstance(node, tuple):
            raise ValueError(f"expected graph.nodes() to be tuple(x,y), but found {node}")
        if not len(node) == 2:
            raise ValueError(f"expected tuples have 2 values, but found {node} (len={len(node)})")
        x, y = node
        if not isinstance(x, (float, int)):
            raise ValueError(f"expected node in graph.nodes() to have (x,y) as float or int, but got {type(x)}")
        if not isinstance(y, (float, int)):
            raise ValueError(f"expected node in graph.nodes() to have (x,y) as float or int, but got {type(y)}")

    plt.figure()
    if nodes:
        xs, ys = [a[0] for a in graph.nodes()], [a[1] for a in graph.nodes()]
        plt.plot(xs, ys)

    if edges:
        for edge in graph.edges():
            s, e, d = edge  # s: (x1,y1), e: (x2,y2), d: distance
            plt.plot([s[0], e[0]], [s[1], e[1]], 'bo-', clip_on=False)

    plt.axis('scaled')
    plt.axis('off')
    return plt


@visualise
def plot_3d(graph, nodes=True, edges=True, rotation='xyz', maintain_aspect_ratio=False):
    """ plots nodes and links using matplotlib3
    :param nodes: bool: plots nodes
    :param edges: bool: plots edges
    :param rotation: str: set view point as one of [xyz,xzy,yxz,yzx,zxy,zyx]
    :param maintain_aspect_ratio: bool: rescales the chart to maintain aspect ratio.
    :return: matlibplot.pyplot
    """
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    if not len(rotation) == 3:
        raise ValueError(f"expected viewpoint as 'xyz' but got: {rotation}")
    for c in 'xyz':
        if c not in rotation:
            raise ValueError(f"rotation was missing {c}.")
    x, y, z = rotation

    # Data for a three-dimensional line
    if edges:
        for edge in graph.edges():
            n1, n2, v = edge
            xyz = dict()
            xyz[x] = [n1[0], n2[0]]
            xyz[y] = [n1[1], n2[1]]
            xyz[z] = [n1[2], n2[2]]
            ax.plot3D(xyz['x'], xyz['y'], xyz['z'], 'gray')

    # Data for three-dimensional scattered points
    if nodes:
        xyz = {x: [], y: [], z: []}
        ix = []
        for idx, node in enumerate(graph.nodes()):
            vx, vy, vz = node  # value of ...
            xyz[x].append(vx)
            xyz[y].append(vy)
            xyz[z].append(vz)
            ix.append(idx)
        ax.scatter3D(xyz['x'], xyz['y'], xyz['z'], c=ix, cmap='Greens')

    if (nodes or edges) and maintain_aspect_ratio:
        nodes = [n for n in graph.nodes()]
        xyz_dir = {'x': 0, 'y': 1, 'z': 2}

        xdim = xyz_dir[x]  # select the x dimension in the projection.
        # as the rotation will change the xdimension index.
        xs = [n[xdim] for n in nodes]  # use the xdim index to obtain the values.
        xmin, xmax = min(xs), max(xs)
        dx = (xmax + xmin) / 2  # determine the midpoint for the dimension.

        ydim = xyz_dir[y]
        ys = [n[ydim] for n in nodes]
        ymin, ymax = min(ys), max(ys)
        dy = (ymax + ymin) / 2

        zdim = xyz_dir[z]
        zs = [n[zdim] for n in nodes]
        zmin, zmax = min(zs), max(zs)
        dz = (zmax + zmin) / 2

        # calculate the radius for the aspect ratio.
        max_dim = max([xmax - xmin, ymax - ymin, zmax - zmin]) / 2

        xa, xb = dx - max_dim, dx + max_dim  # lower, uppper
        ax.set_xlim(xa, xb)  # apply the lower and upper to the axis.
        ya, yb = dy - max_dim, dy + max_dim
        ax.set_ylim(ya, yb)
        za, zb = dz - max_dim, dz + max_dim
        ax.set_zlim(za, zb)

    ax.set_xlabel(f'{x} Label')
    ax.set_ylabel(f'{y} Label')
    ax.set_zlabel(f'{z} Label')
    return plt



from time import process_time
from graph import Graph
from bisect import insort
from itertools import product


__description__ = """
We've decided to refer to the optimisation problem of finding
the fewest number of moves that resolve a traffic jam as
a traffic scheduling problem.
"""


class UnSolvable(ValueError):
    """ Value Error raised if the inputs have no solution. """
    pass


class NoSolution(ValueError):
    """ Value Error raised if the method could not identify a valid solution """
    pass


class StopCondition(Exception):
    """ exception used to stop the search """
    pass


class Timer(object):
    def __init__(self, timeout=None):
        """
        :param timeout: int/float in milliseconds.
        """
        if timeout is None:
            timeout = float('inf')
        if not isinstance(timeout, (float, int)):
            raise ValueError(f"timeout is {type(timeout)} not int or float > 0")
        if timeout < 0:
            raise ValueError(f"timeout must be >0, but was {timeout}")

        self.limit = timeout
        self.start = process_time()
        self._counter = 0
        self._expired = False

    def expired(self):
        """ returns bool"""
        if self._expired:
            return True

        # we use a counter as there is no reason to check the time 100_000 times/second.
        if self._counter > 0:
            self._counter -= 1
            return False
        # The counter was zero, so now we check the time and reset the counter.
        self._counter = 100
        if process_time() - self.start > (self.limit / 1000):
            self._expired = True
        return False


class Load(object):
    _empty = frozenset()

    __slots__ = ["id", "start", "ends", "prohibited"]

    def __init__(self, id, start, ends=None, prohibited=None):
        """
        :param id: unique load reference
        :param start: start node
        :param ends: end node(s)
            - list, set, frozenset, tuples are interpreted as a number of candidate destinations.
            - all other types are interpreted as 1 destination.
        :param prohibited: iterable with nodes that cannot be in the solution.
        """
        self.id = id
        self.start = start
        if isinstance(ends, frozenset):
            if not ends:
                raise ValueError(f"end is an empty {type(ends)}?")
            self.ends = ends
        elif isinstance(ends, (list, set, tuple)):
            if not ends:
                raise ValueError(f"end is an empty {type(ends)}?")
            self.ends = frozenset(ends)
        elif ends is None:
            self.ends = frozenset([start])
        else:
            self.ends = frozenset([ends])
        assert isinstance(self.ends, frozenset), "bad logic!"

        if prohibited is not None:
            if not isinstance(prohibited, (list, frozenset, set, tuple)):
                raise TypeError(f"Got {type(prohibited)}, expected list, set or tuple")
            self.prohibited = frozenset(prohibited)
        else:
            self.prohibited = Load._empty  # link to an empty frozen set.

    def __str__(self):
        return f"Load(id={self.id}, start={self.start}, end={self.ends}, prohibited={self.prohibited})"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if not isinstance(other, Load):
            raise TypeError
        return all([
            self.start == other.start,
            self.ends == other.ends,
            self.prohibited == other.prohibited
        ])


# helpers
def check_user_input(graph, loads):
    """ checks the user inputs to be valid.

    :param graph network available for routing.
    :param loads: dictionary or list with load id and preferred route. Examples:

    loads_as_list = [
        {'id': 1, 'start': 1, 'end': 3},  # keyword prohibited is missing.
        {'id': 2, 'start': 2, 'end': [3, 4, 5], 'prohibited': [7, 8, 9]},
        {'id': 3, 'start': 3, 'end': [4, 5], 'prohibited': [2]}  # gateway to off limits.
        {'id': 4, 'start': 8}  # load 4 is where it needs to be.
    ]

    loads_as_dict = {
        1: (1, 3),  # start, end, None
        2: (2, [3, 4, 5], [7, 8, 9]),  # start, end(s), prohibited
        3: (3, [4, 5], [2]),
        4: (8, ),
    }

    returns: list of Loads
    """
    if not isinstance(graph, Graph):
        raise TypeError(f"expected graph, not {type(graph)}")

    all_loads = {}
    all_nodes = set()
    if isinstance(loads, list):
        for d in loads:
            if not isinstance(d, dict):
                raise TypeError(f"Got {type(d)}, expected dict.")
            L = Load(**d)
            all_loads[L.id] = L
            all_nodes.add(L.start)
            all_nodes.update(L.ends)
    elif isinstance(loads, dict):
        for i, t in loads.items():
            if not isinstance(t, (tuple, list)):
                raise TypeError(f"Got {type(t)}, expected tuple")
            L = Load(i, *t)
            all_loads[L.id] = L
            all_nodes.add(L.start)
            all_nodes.update(L.ends)
    else:
        raise TypeError("loads not recognised. Please see docstring.")

    if not all_nodes.issubset(set(graph.nodes())):  # then something is wrong. Let's analyze to help the programmer...
        diff = all_nodes.difference(set(graph.nodes()))
        for load in all_loads.values():
            if load.start in diff:
                raise ValueError(f"Load {load.id}'s start ({load.start}) is not in the graph.")
            if load.ends.intersection(diff):
                raise ValueError(f"Load {load.id}'s ends ({load.ends.intersection(diff)}) is/are not in the graph.")
            if load.prohibited.intersection(diff):
                raise ValueError(f"Load {load.id}'s prohibited node(s) ({load.prohibited.intersection(diff)}) is/are not in the graph.")

    assignment_options = {load.id: frozenset(load.ends) for load in all_loads.values() }
    if not is_ap_solvable(assignment_options):
        raise UnSolvable(f"There are not enough ends for all the loads to be assigned to a destination.")

    for load in all_loads.values():
        if load.prohibited:
            gc = graph.copy()
            for n in load.prohibited:
                gc.del_node(n)
        else:
            gc = graph

        for node in gc.breadth_first_walk(start=load.start):
            if node in load.ends:
                break
        else:
            ends = f"any {tuple(load.ends)}" if len(load.ends)>1 else f"{list(load.ends)[0]}"
            raise UnSolvable(f"load {load.id} has no path from {load.start} to {ends}")
    return all_loads


def is_ap_solvable(assignments):
    """
    A number of loads need to be assigned to a destination.
    The loads have preferences, f.x.

        A = {1,2}, B = {2,3}, C = {1,2,3}  # solveable
        A = {1,2}, B = {1,3}, C = {1}      # solveable
        A = {1,2}, B = {1,2}, C = {1}      # not solveable.

    This method checks if the assignment is possible
    """
    if not isinstance(assignments, dict):
        raise TypeError
    if not all(isinstance(i, (frozenset, set)) for i in assignments.values()):
        raise TypeError

    all_ends = set().union(*assignments.values())

    assignment = {}

    for load_id, ends in sorted(assignments.items(), key=lambda x: len(x[-1])):
        options = set(ends).intersection(all_ends)
        if not options:
            return False
        selection = options.pop()
        all_ends.remove(selection)
        assignment[load_id] = selection
    return True


def path_to_moves(path):
    """
    translate path with states into a motion sequence.
    :param path: list of tuples with [(load id, location), ... ]
    """
    moves = []
    s1 = path[0]
    for s2 in path[1:]:
        for p1, p2 in zip(s1, s2):
            if p1 != p2:
                moves.append({p1[0]: (p1[1], p2[1])})  # {load id: (location1, location2)}
        s1 = s2
    return moves


def moves_to_synchronous_moves(moves, loads):
    """ translates the list of moves returned from the traffic jam solver to a list
    of moves that can be made concurrently.

    :param moves: list of loads and moves, e.g. [{1: (2,3)}, {2:(1,2)}, ... ]
    :param loads: dict with loads and paths, e.g. {1: [2,3,4], 2: [1,2,3], ... }
    :return: list of synchronous loads and moves, e.g. [{1:(2,3), {2:(1,2}}, {1:(3,4), 2:(2,3)}, ...]
    """
    moves = [(k,) + v for move in moves for k, v in move.items()]  # create independent copy
    assert isinstance(loads, dict)
    assert all(isinstance(ld, Load) for ld in loads.values())

    occupied_locations = {L.start for L in loads.values()}  # loads are required in case that a load doesn't move.
    synchronous = []

    while moves:
        current_moves = {}
        for move in moves[:]:
            load, n1, n2 = move
            if load in current_moves:
                break
            if n2 in occupied_locations:
                continue
            current_moves[load] = (n1, n2)
            occupied_locations.remove(n1)
            occupied_locations.add(n2)
            moves.remove(move)
        synchronous.append(current_moves)
    return synchronous


class State(object):
    def __init__(self, loads, distance=0, gradient=float('inf')):
        self.loads = loads
        self._hash = hash(self.loads)  # by pre-calculating the hash, we save cpu time later.
        self.distance = distance
        self.gradient = gradient
        self._d = None  # for most cases this isn't needed.

    def __str__(self):
        return f"State({self.loads}, {self.distance})"

    def __repr__(self):
        return str(self)

    def __lt__(self, other):
        if self.gradient == other.gradient:
            return self.distance < other.distance
        else:
            return self.gradient < other.gradient

    def __iter__(self):
        for lid, loc in self.loads:
            yield lid, loc

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        # this is actually required for dict's to work. For more see example on:
        # https://stackoverflow.com/questions/9010222/why-can-a-python-dict-have-multiple-keys-with-the-same-hash?noredirect=1&lq=1
        return self._hash == other._hash

    def occupied(self):
        return {i[1] for i in self.loads}


class JamSolver(object):
    def __init__(self, graph, loads, timer):
        if not isinstance(graph, Graph):
            raise TypeError
        if not isinstance(loads, dict):
            raise TypeError
        if not all(isinstance(i, Load) for i in loads.values()):
            raise TypeError
        if not all(lid == load.id for lid, load in loads.items()):
            raise ValueError
        if not isinstance(timer, Timer):
            raise TypeError

        self.graph = graph
        self.movements = Graph()
        self.loads = loads
        self.timer = timer

        initial_state = State(loads=tuple((ld.id, ld.start) for ld in self.loads.values()), distance=0)
        self.start = initial_state
        self.ends = set()

        self.movements.add_node(initial_state)

        self.forward_queue = [initial_state]                # this is the working queue with priority.
        self.forward_edge = {initial_state: initial_state}  # this is a duplicate of items in the work queue
        self.forward_visited = set()                        # we don't want to spend CPU time on this.

        self.reverse_queue = []
        self.reverse_edge = {}
        self.reverse_visited = set()

        self.max_search_distance = float('inf')

        self.final_states = set()

        for state in self._end_state_gen():
            self.ends.add(state)
            self.reverse_queue.append(state)
            self.reverse_edge[state] = state
            self.final_states.add(state)

        self.distance_maps = {}
        # The distance maps are used as proxy for simulated annealing.
        # The closer a load is to any of it's destinations, the lower the temperature.
        # when all loads have reach a destination, the temperature is zero.
        for load_id, load in self.loads.items():
            self.distance_maps[load_id] = self.graph.distance_map(ends=load.ends, reverse=True)

        self.done = False
        self.return_on_first = False

    def __str__(self):
        s = "timed out" if self.timer.expired() else "running"
        return f"<{self.__class__.__name__}> {s}"

    def _distance(self, load_id, location):
        try:
            return self.distance_maps[load_id][location]
        except KeyError:
            return max(self.distance_maps[load_id].values()) + 1

    def solve(self, return_on_first=False):
        if not isinstance(return_on_first, bool):
            raise TypeError
        self.return_on_first = return_on_first

        try:
            self._search()
        except StopCondition as e:
            print(str(e))
        return self._shortest_path_multiple_ends()

    def _end_state_gen(self):
        ids, destinations = [], []
        for load in self.loads.values():
            ids.append(load.id)
            destinations.append(list(load.ends))

        for combo in product(*destinations):
            if len(set(combo)) != len(destinations):
                continue  # it's a duplicate.
            state = tuple((lid, loc) for lid, loc in zip(ids, combo))
            yield State(state)

    def _search(self):
        while not self.timer.expired():

            self._find_forward_options()  # forward search
            self._find_reverse_options()  # reverse search

            if not any((self.forward_queue, self.reverse_queue)):
                raise StopCondition("queue exhausted")
                # note: This may mean the solution-landscape is exhausted. Not that something is wrong.

    def _find_forward_options(self):
        if not self.forward_queue:
            return
        state = self.forward_queue.pop(0)
        self.forward_edge.pop(state)
        self.forward_visited.add(state)

        occupied = state.occupied()
        for load_id, location in state:
            load = self.loads[load_id]

            options = sorted((d, e) for s, e, d in self.graph.edges(from_node=location)
                             if e not in occupied and e not in load.prohibited)

            for distance, option in options:
                new_distance_traveled = state.distance + distance

                loads = tuple((lid, loc) if lid != load_id else (load_id, option) for lid, loc in state)
                gradient = sum((self._distance(lid, loc) for lid, loc in loads))
                new_state = State(loads, new_distance_traveled, gradient)

                if self.movements.edge(state, new_state, float('inf')) < distance:
                    continue
                else:
                    self.movements.add_edge(state, new_state, distance)

                if new_state in self.forward_edge:
                    existing_new_state = self.forward_edge.get(new_state)
                    if new_state < existing_new_state:  # replace!
                        self.forward_edge[new_state] = new_state
                        self.forward_queue.remove(existing_new_state)
                        insort(self.forward_queue, new_state)
                    else:
                        pass
                    continue  # case is already queued for testing.
                elif new_state in self.forward_visited or new_state in self.reverse_visited:
                    continue  # seen before
                else:
                    self.forward_edge[new_state] = new_state
                    insort(self.forward_queue, new_state)

                if new_state in self.reverse_visited:
                    self._match()

    def _shortest_path_multiple_ends(self):
        """ helper that identifies the fewest number of moves required to reach any
        of the valid end states """
        d_min, p_min = float('inf'), None
        for end in self.final_states:
            if end in self.movements:
                d, p = self.movements.shortest_path(self.start, end)
                if d < d_min:  # then this solution is better than the previous.
                    d_min = d
                    p_min = p
        return d_min, p_min

    def _match(self):
        """ helper that updates the search queues when one finds a matching state in the other."""
        d, p = self._shortest_path_multiple_ends()
        self.max_search_distance = min(d, self.max_search_distance)

        self.forward_visited.update({s for s in self.forward_queue if s.distance > self.max_search_distance})
        self.forward_queue = [s for s in self.forward_queue if s.distance <= self.max_search_distance]
        self.forward_edge = {s: s for s in self.forward_queue}

        self.reverse_visited.update({s for s in self.reverse_queue if s.distance > self.max_search_distance})
        self.reverse_queue = [s for s in self.reverse_queue if s.distance <= self.max_search_distance]
        self.reverse_edge = {s: s for s in self.reverse_queue}

        if self.return_on_first:
            raise StopCondition("solution found")

    def _find_reverse_options(self):
        if not self.reverse_queue:  # backward...
            return
        state = self.reverse_queue.pop(0)
        self.reverse_edge.pop(state)
        self.reverse_visited.add(state)

        occupied = state.occupied()
        for load_id, location in state:
            load = self.loads[load_id]

            options = sorted((d, s) for s, e, d in self.graph.edges(to_node=location) if
                             s not in occupied and s not in load.prohibited)

            for distance, option in options:
                new_distance_traveled = state.distance + distance

                loads = tuple((lid, loc) if lid != load_id else (load_id, option) for lid, loc in state)
                gradient = sum((self._distance(lid, loc) for lid, loc in loads))
                new_state = State(loads, new_distance_traveled, gradient)

                if self.movements.edge(new_state, state, float('inf')) < distance:
                    continue
                else:
                    self.movements.add_edge(new_state, state, distance)

                if new_state in self.reverse_edge:
                    existing_new_state = self.reverse_edge.get(new_state)
                    if new_state < existing_new_state:  # replace!
                        self.reverse_edge[new_state] = new_state
                        self.reverse_queue.remove(existing_new_state)
                        insort(self.reverse_queue, new_state)
                    else:
                        pass
                    continue  # case is already queued for testing.
                elif new_state in self.reverse_visited or new_state in self.forward_visited:
                    continue  # seen before
                else:
                    self.reverse_edge[new_state] = new_state
                    insort(self.reverse_queue, new_state)

                if new_state in self.forward_visited:
                    self._match()


def jam_solver(graph, loads, timeout=None, synchronous_moves=True, return_on_first=False):
    """
    The traffic jam solver
    - a bidirectional search algorithm that uses simulated annealing
    to induce bias to accelerate the solvers direction of search.

    :param graph network available for routing.
    :param loads: dictionary or list with load id and preferred route. Examples:

    loads_as_list = [
        {'id': 1, 'start': 1, 'end': 3},  # keyword prohibited is missing.
        {'id': 2, 'start': 2, 'end': [3, 4, 5], 'prohibited': [7, 8, 9]},
        {'id': 3, 'start': 3, 'end': [4, 5], 'prohibited': [2]}  # gateway to off limits.
        {'id': 4, 'start': 8}  # load 4 is where it needs to be.
    ]

    loads_as_dict = {
        1: (1, 3),  # start, end, None
        2: (2, [3, 4, 5], [7, 8, 9]),  # start, end(s), prohibited
        3: (3, [4, 5], [2]),
        4: (8, ),
    }

    :param timeout: None or Number of milliseconds.
    :param synchronous_moves: boolean, set to True to return concurrent moves
    :param return_on_first: boolean, tell solver to stop at first valid solution,
                            disregarding whether the solution is the most energy
                            efficient.
    :return: list of dictionaries as sequence of moves. Example:

    solution = [{'b': (5, 6), 'c': (4, 5), 'e': (1, 4)},              # 1st moves.
                {'b': (6, 3), 'c': (5, 6), 'e': (4, 5), 'a': (2, 1)}, # 2nd moves.
                {'b': (3, 2), 'c': (6, 3), 'e': (5, 6)},              # 3rd moves.
                {'e': (6, 9)}]                                        # 4th move.

    """
    all_loads = check_user_input(graph, loads)
    timer = Timer(timeout)
    c = JamSolver(graph, all_loads, timer)
    d, p = c.solve(return_on_first)

    if not p:
        if timer.expired():
            raise UnSolvable(f"no solution found with timeout = {timeout} msecs")
        else:
            raise NoSolution(f"no solution found.")

    moves = path_to_moves(p)
    if synchronous_moves:
        return moves_to_synchronous_moves(moves, all_loads)
    return moves




"""
graph-theory
"""
build_tag = "5f1ab7d5047312cd2c4eda6ce6a98d123a90b9a81ea484eb2741e3a29891144a"
from setuptools import setup
from pathlib import Path


folder = Path(__file__).parent
file = "README.md"
readme = folder / file
assert isinstance(readme, Path)
assert readme.exists(), readme
with open(str(readme), encoding='utf-8') as f:
    long_description = f.read()

keywords = list({
    'complex-networks', 'discrete mathematics', 'graph', 'Graph Theory', 'graph-algorithms', 'graph-analysis',
    'graph-generation', 'graph-theory', 'graph-visualization', 'graphs', 'math', 'Mathematics', 'maths',
    'minimum-spanning-trees', 'network', 'Networks', 'optimization', 'python', 'shortest-path', 'tsp', 'tsp-solver',
    'assignment problem', 'flow-problem', 'hash', 'graph-hash', 'random graph', 'search', 'cycle', 'path',
    'component', 'components', 'adjacency', 'matrix', 'all pairs shortest path', 'finite state machine', 'fsm',
    'traffic-jam', 'traffic-jam-solver', 'solver', 'hill-climbing', 'simple', 'simple-path', 'critical', 'path',
    'method', 'critical-path', 'minimize', 'minimise', 'optimize', 'optimise', 'merkle', 'tree', 'merkle-tree',
    'hash-tree'
})

keywords.sort(key=lambda x: x.lower())


setup(
    name="graph-theory",
    version="2022.3.9.54615",
    url="https://github.com/root-11/graph-theory",
    license="MIT",
    author="Bjorn Madsen",
    author_email="bjorn.madsen@operationsresearchgroup.com",
    description="A graph library",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=keywords,
    packages=["graph"],
    include_package_data=True,
    data_files=[(".", ["LICENSE", "README.md"])],
    platforms="any",
    install_requires=[],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)





from collections import defaultdict, deque
from collections.abc import Iterable
from heapq import heappop, heappush
from itertools import combinations, chain
from bisect import insort

from graph.visuals import plot_3d

__description__ = """
The graph-theory library is organised in the following way for clarity of structure:

1. BasicGraph (class) - with general methods for all subclasses.
2. All methods for class Graph in same order as on Graph.
3. Graph (class)
4. Graph3D (class) 
"""


class BasicGraph(object):
    """
    BasicGraph is the base graph that all methods use.

    For methods, please see the documentation on the
    individual functions, by importing them separately.
    """

    def __init__(self, from_dict=None, from_list=None):
        """
        :param from_dict: creates graph for dictionary {n1:{n2:d} ...
        :param from_list: creates graph from list of edges(n1,n2,d)
        """
        self._nodes = {}
        self._edges = defaultdict(dict)
        self._edge_count = 0
        self._reverse_edges = defaultdict(dict)
        self._in_degree = defaultdict(int)
        self._out_degree = defaultdict(int)

        if from_dict is not None:
            self.from_dict(from_dict)
        elif from_list is not None:
            self.from_list(from_list)

    def __str__(self):
        return f"{self.__class__.__name__}({len(self._nodes)} nodes, {self._edge_count} edges)"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self._nodes != other._nodes:
            return False
        if self._edges != other._edges:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getitem__(self, item):
        raise ValueError("Use g.node(n1) or g.edge(n1,n2)")

    def __setitem__(self, key, value):
        raise ValueError("Use add_edge(node1, node2, value)")

    def __delitem__(self, key):
        raise ValueError("Use del_edge(node1, node2)")

    def __contains__(self, item):
        """
        :returns bool: True if node in Graph.
        """
        return item in self._nodes

    def __len__(self):
        raise ValueError("Use len(g.nodes()) or len(g.edges())")

    def copy(self):
        g = Graph()
        for n in self._nodes:
            g.add_node(n, obj=self._nodes[n])
        for s, e, d in self.edges():
            g.add_edge(s, e, d)
        return g

    def add_edge(self, node1, node2, value=1, bidirectional=False):
        """
        :param node1: hashable node
        :param node2: hashable node
        :param value: numeric value (int or float)
        :param bidirectional: boolean.
        """
        if isinstance(value, (dict, list, tuple)):
            raise ValueError("value cannot be {}".format(type(value)))
        if node1 not in self._nodes:
            self.add_node(node1)
        if node2 not in self._nodes:
            self.add_node(node2)

        if node1 in self._edges and node2 in self._edges[node1]:  # it's a value update.
            self._edges[node1][node2] = value
            self._reverse_edges[node2][node1] = value
        else:  # it's a new edge.
            self._edges[node1][node2] = value
            self._reverse_edges[node2][node1] = value
            self._out_degree[node1] += 1
            self._in_degree[node2] += 1
            self._edge_count += 1

        if bidirectional:
            self.add_edge(node2, node1, value, bidirectional=False)

    def edge(self, node1, node2, default=None):
        """Retrieves the edge (node1, node2)

        Alias for g[node1][node2]

        :param node1: node id
        :param node2: node id
        :param default: returned value if edge doesn't exist.
        :return: edge(node1,node2)
        """
        try:
            return self._edges[node1][node2]
        except KeyError:
            return default

    def reverse_edge(self, node2, node1, default=None):
        """ retrieves the edge from node2 to node1 """
        try:
            return self._reverse_edges[node2][node1]
        except KeyError:
            return default

    def del_edge(self, node1, node2):
        """
        removes edge from node1 to node2
        :param node1: node
        :param node2: node
        """
        try:
            del self._edges[node1][node2]
        except KeyError:
            return

        del self._reverse_edges[node2][node1]
        self._edge_count -= 1

        if self._out_degree[node1] == 0:
            assert False
        self._out_degree[node1] -= 1

        if self._in_degree[node2] == 0:
            assert False
        self._in_degree[node2] -= 1

    def add_node(self, node_id, obj=None):
        """
        :param node_id: any hashable node.
        :param obj: any object that the node should refer to.

        PRO TIP: To retrieve the node obj use g.node(node_id)

        """
        if node_id in self._nodes:  # it's an object update.
            self._nodes[node_id] = obj
        else:
            self._nodes[node_id] = obj
            self._in_degree[node_id] = 0
            self._out_degree[node_id] = 0

    def node(self, node_id):
        """
        Retrieves the node object

        :param node_id: id of node in graph.
        :return: node object
        """
        return self._nodes.get(node_id, None)

    def del_node(self, node_id):
        """
        Deletes the node and all its connections.
        :param node_id: node_id
        :return: None
        """
        if node_id not in self._nodes:
            return

        # outgoing
        for n2, d in self._edges[node_id].copy().items():
            self.del_edge(node_id, n2)

        # incoming
        for n1, d in self._reverse_edges[node_id].copy().items():
            self.del_edge(n1, node_id)

        del self._edges[node_id]
        del self._reverse_edges[node_id]  # removes outgoing edges
        del self._nodes[node_id]
        del self._in_degree[node_id]
        del self._out_degree[node_id]

    def nodes(self,
              from_node=None, to_node=None,
              in_degree=None, out_degree=None):
        """
        :param from_node (optional) return nodes with edges from 'from_node'
        :param to_node (optional) returns nodes with edges into 'to_node'
        :param in_degree (optional) returns nodes with in_degree=N
        :param out_degree (optional) returns nodes with out_degree=N

        :return list of node ids.
        """
        inputs = sum([1 for i in (from_node, to_node, in_degree, out_degree) if i is not None])
        if inputs > 1:
            m = []
            a = (from_node, to_node, in_degree, out_degree)
            b = ("from_node", "to_node", "in_degree", "out_degree")
            for i in zip(a, b):
                if i is not None:
                    m.append("{}={}".format(b, a))
            raise ValueError("nodes({}) has too many inputs. Pick one.".format(m))

        if inputs == 0:
            return list(self._nodes.keys())

        if from_node is not None:
            if from_node in self._edges:
                return list(self._edges[from_node])
            return []

        if to_node is not None:
            return list(self._reverse_edges[to_node])

        if in_degree is not None:
            if not isinstance(in_degree, int) or in_degree < 0:
                raise ValueError("in_degree must be int >= 0")
            return [n for n, cnt in self._in_degree.items() if cnt == in_degree]

        if out_degree is not None:
            if not isinstance(out_degree, int) or out_degree < 0:
                raise ValueError("out_degree must be int >= 0")
            return [n for n, cnt in self._out_degree.items() if cnt == out_degree]

    def edges(self, path=None, from_node=None, to_node=None):
        """
        :param path (optional) list of nodes for which the edges are wanted.
        :param from_node (optional) for which outgoing edges are returned.
        :param to_node (optional) for which incoming edges are returned.
        :return list of edges (n1, n2, value)
        """
        inputs = sum([1 for i in (from_node, to_node, path) if i is not None])
        if inputs > 1:
            m = []
            a = (path, from_node, to_node)
            b = ("path", "from_node", "to_node")
            for i in zip(a, b):
                if i is not None:
                    m.append("{}={}".format(b, a))
            raise ValueError("edges({}) has too many inputs. Pick one.".format(m))

        if path:
            if not isinstance(path, list):
                raise ValueError("expects a list")
            if len(path) < 2:
                raise ValueError("path of length 1 is not a path.")

            return [(path[ix], path[ix + 1], self._edges[path[ix]][path[ix + 1]])
                    for ix in range(len(path) - 1)]

        if from_node:
            if from_node in self._edges:
                return [(from_node, n2, cost) for n2, cost in self._edges[from_node].items()]
            else:
                return []

        if to_node:
            if to_node in self._reverse_edges:
                return [(n1, to_node, value) for n1, value in self._reverse_edges[to_node].items()]
            else:
                return []

        return [(n1, n2, out[n2]) for n1, out in self._edges.items() for n2 in out]

    def from_dict(self, dictionary):
        """
        Updates the graph from dictionary
        :param dictionary:

        d = {1: {2: 10, 3: 5},
             2: {4: 1, 3: 2},
             3: {2: 3, 4: 9, 5: 2},
             4: {5: 4},
             5: {1: 7, 4: 6}}

        G = Graph(from_dict=d)

        :return: None
        """
        if not isinstance(dictionary, dict):
            raise TypeError(f"expected dict, not {type(dictionary)}")
        for n1, e in dictionary.items():
            if not e:
                self.add_node(n1)
            else:
                for n2, v in e.items():
                    self.add_edge(n1, n2, v)

    def to_dict(self):
        """ creates a nested dictionary from the graph.
        :return dict d[n1][n2] = distance
        """
        d = {n: {} for n in self.nodes()}
        for n1, n2, dist in self.edges():
            d[n1][n2] = dist
        return d

    def from_list(self, links):
        """
        updates the graph from a list of links.
        :param links: list

        links = [
            (1, 2, 18),
            (1, 3, 10),
            (2, 4, 7),
            (2, 5, 6),
            (3, 4, 2),
            (11,)      # node with no links.
        ]
        """
        if not isinstance(links, Iterable):
            raise TypeError(f"Expected iterable, not {type(links)}")
        for item in links:
            assert isinstance(item, (list, tuple))
            if len(item) > 1:
                self.add_edge(*item)
            else:
                self.add_node(item[0])

    def to_list(self):
        """ returns list of edges and nodes."""
        return self.edges() + [(i,) for i in self.nodes()]

    def is_connected(self, n1, n2):
        """ helper determining if two nodes are connected using BFS. """
        if n1 in self._edges:
            q = deque([n1])
            visited = set()
            while q:
                n = q.popleft()
                if n not in visited:
                    visited.add(n)
                for c in self._edges[n]:
                    if c == n2:
                        return True  # <-- Exit if connected.
                    if c in visited:
                        continue
                    else:
                        q.append(c)
        return False  # <-- Exit if not connected.

    def in_degree(self, node):
        """ returns the number of edges incoming on a node """
        return self._in_degree[node]

    def out_degree(self, node):
        """ returns the number of edges departing from a node"""
        return self._out_degree[node]


# Graph functions
# -----------------------------
def shortest_path(graph, start, end, avoids=None):
    """ single source shortest path algorithm.
    :param graph: class Graph
    :param start: start node
    :param end: end node
    :param avoids: optional set,frozenset or list of nodes that cannot be a part of the path.
    :return distance, path (as list),
            returns float('inf'), [] if no path exists.
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError(f"{start} not in graph")
    if end not in graph:
        raise ValueError(f"{end} not in graph")
    if avoids is None:
        visited = set()
    elif not isinstance(avoids, (frozenset, set, list)):
        raise TypeError(f"Expect obstacles as set or frozenset, not {type(avoids)}")
    else:
        visited = set(avoids)

    q, minimums = [(0, 0, start, ())], {start: 0}
    i = 1
    while q:
        (cost, _, v1, path) = heappop(q)
        if v1 not in visited:
            visited.add(v1)
            path = (v1, path)

            if v1 == end:  # exit criteria.
                L = []
                while path:
                    v, path = path[0], path[1]
                    L.append(v)
                L.reverse()
                return cost, L

            for _, v2, dist in graph.edges(from_node=v1):
                if v2 in visited:
                    continue
                prev = minimums.get(v2, None)
                next_node = cost + dist
                if prev is None or next_node < prev:
                    minimums[v2] = next_node
                    heappush(q, (next_node, i, v2, path))
                    i += 1
    return float("inf"), []


def breadth_first_search(graph, start, end):
    """ Determines the path from start to end with fewest nodes.
    :param graph: class Graph
    :param start: start node
    :param end: end node
    :return: path
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError(f"{start} not in graph")
    if end not in graph:
        raise ValueError(f"{end} not in graph")

    visited = {start: None}
    q = deque([start])
    while q:
        node = q.popleft()
        if node == end:
            path = deque()
            while node is not None:
                path.appendleft(node)
                node = visited[node]
            return list(path)
        for next_node in graph.nodes(from_node=node):
            if next_node not in visited:
                visited[next_node] = node
                q.append(next_node)
    return []


def breadth_first_walk(graph, start, end=None, reversed_walk=False):
    """
    :param graph: Graph
    :param start: start node
    :param end: end node.
    :param reversed_walk: if True, the BFS traverse the graph backwards.
    :return: generator for walk.

    To walk all nodes use: `[n for n in g.breadth_first_walk(start)]`
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError(f"{start} not in graph")
    if end is not None and end not in graph:
        raise ValueError(f"{end} not in graph. Use `end=None` if you want exhaustive search.")
    if not isinstance(reversed_walk, bool):
        raise TypeError(f"reversed_walk should be boolean, not {type(reversed_walk)}: {reversed_walk}")

    visited = {start: None}
    q = deque([start])
    while q:
        node = q.popleft()
        yield node
        if node == end:
            break
        L = graph.nodes(from_node=node) if not reversed_walk else graph.nodes(to_node=node)
        for next_node in L:
            if next_node not in visited:
                visited[next_node] = node
                q.append(next_node)


def distance_map(graph, starts=None, ends=None, reverse=False):
    """ Maps the shortest path distance from any start to any end.
    :param graph: instance of Graph
    :param starts: None, node or (set,list,tuple) of nodes
    :param ends: None (exhaustive map), node or (set,list,tuple) of nodes
                 that terminate the search when all are found
    :param reverse: bool: walks the map from the ends towards the starts using reversed edges.
    :return: dictionary with {node: distance from start}
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    if not isinstance(reverse, bool):
        raise TypeError("keyword reverse was not boolean")

    if all((starts is None, ends is None)):
        raise ValueError("starts and ends cannot both be None")
    if all((starts is None, reverse is False)):
        raise ValueError("walking forward from end doesn't make sense.")
    if all((ends is None, reverse is True)):
        raise ValueError("walking from reverse from start doesn't make sense.")

    if isinstance(starts, Iterable):
        starts = set(starts)
    else:
        starts = {starts}
    if any(start not in graph for start in starts if start is not None):
        missing = [start not in graph for start in starts]
        raise ValueError(f"starts: ({missing}) not in graph")

    if isinstance(ends, Iterable):
        ends = set(ends)
    else:
        ends = {ends}
    if any(end not in graph for end in ends if end is not None):
        missing = [end not in graph for end in ends if end is not None]
        raise ValueError(f"{missing} not in graph. Use `end=None` if you want exhaustive search.")

    if not reverse:
        ends_found = set()
        visited = {start: 0 for start in starts}
        q = deque(starts)
        while q:
            if ends_found == ends:
                break
            n1 = q.popleft()
            if n1 in ends:
                ends_found.add(n1)
            d1 = visited[n1]
            for _, n2, d in graph.edges(from_node=n1):
                if n2 not in visited:
                    q.append(n2)
                visited[n2] = min(d1+d, visited.get(n2, float('inf')))

    else:
        starts_found = set()
        visited = {end: 0 for end in ends}
        q = deque(ends)
        while q:
            if starts_found == starts:
                break
            n2 = q.popleft()
            if n2 in starts:
                starts_found.add(n2)
            d2 = visited[n2]
            for n1, _, d in graph.edges(to_node=n2):
                if n1 not in visited:
                    q.append(n1)
                visited[n1] = min(d+d2, visited.get(n1, float('inf')))

    return visited


def depth_first_search(graph, start, end):
    """
    Determines path from start to end using
    'depth first search' with backtracking.

    :param graph: class Graph
    :param start: start node
    :param end: end node
    :return: path as list of nodes.
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError(f"{start} not in graph")
    if end not in graph:
        raise ValueError(f"{end} not in graph")

    q = deque([start])  # q = [start]
    # using deque as popleft and appendleft is faster than using lists. For details see
    # https://stackoverflow.com/questions/23487307/python-deque-vs-list-performance-comparison
    path = []
    visited = set()
    while q:
        n1 = q.popleft()  # n1 = q.pop()
        visited.add(n1)
        path.append(n1)
        if n1 == end:
            return path  # <-- exit if end is found.
        for n2 in graph.nodes(from_node=n1):
            if n2 in visited:
                continue
            q.appendleft(n2)   # q.append(n2)
            break
        else:
            path.remove(n1)
            while not q and path:
                for n2 in graph.nodes(from_node=path[-1]):
                    if n2 in visited:
                        continue
                    q.appendleft(n2)  # q.append(n2)
                    break
                else:
                    path = path[:-1]
    return None  # <-- exit if not path was found.


def depth_scan(graph, start, criteria):
    """ traverses the descendants of node `start` using callable `criteria` to determine
    whether to terminate search along each branch in `graph`.

    :param graph: class Graph
    :param start: start node
    :param criteria: function to terminate scan along a branch must return bool
    :return: set of nodes
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError(f"{start} not in graph")
    if not callable(criteria):
        raise TypeError(f"Expected {criteria} to be callable")
    if not criteria(start):
        return set()

    q = [start]
    path = []
    visited = set()
    while q:
        n1 = q.pop()
        visited.add(n1)
        path.append(n1)
        for n2 in graph.nodes(from_node=n1):
            if n2 in visited:
                continue
            if not criteria(n2):
                visited.add(n2)
                continue
            q.append(n2)
            break
        else:
            path.remove(n1)
            while not q and path:
                for n2 in graph.nodes(from_node=path[-1]):
                    if n2 in visited:
                        continue
                    if not criteria(n2):
                        visited.add(n2)
                        continue
                    q.append(n2)
                    break
                else:
                    path = path[:-1]
    return visited


def distance(graph, path):
    """ Calculates the distance for the path in graph
    :param graph: class Graph
    :param path: list of nodes
    :return: distance
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if not isinstance(path, (tuple, list)):
        raise TypeError(f"expected tuple or list, not {type(path)}")

    cache = defaultdict(dict)
    path_length = 0
    for idx in range(len(path) - 1):
        n1, n2 = path[idx], path[idx + 1]

        # if the edge exists...
        d = graph.edge(n1, n2, default=None)
        if d:
            path_length += d
            continue

        # if we've seen the edge before...
        d = cache.get((n1, n2), None)
        if d:
            path_length += d
            continue

        # if there no alternative ... (search)
        d, _ = shortest_path(graph, n1, n2)
        if d == float('inf'):
            return float('inf')  # <-- Exit if there's no path.
        else:
            cache[(n1, n2)] = d
        path_length += d
    return path_length


def maximum_flow(graph, start, end):
    """
    Returns the maximum flow graph
    :param graph: instance of Graph
    :param start: node
    :param end: node
    :return: flow, graph
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError(f"{start} not in graph")
    if end not in graph:
        raise ValueError(f"{end} not in graph")

    inflow = sum(d for s, e, d in graph.edges(from_node=start))
    outflow = sum(d for s, e, d in graph.edges(to_node=end))
    unassigned_flow = min(inflow, outflow)  # search in excess of this 'flow' is a waste of time.
    total_flow = 0
    # -----------------------------------------------------------------------
    # The algorithm
    # I reviewed a number of algorithms, such as Ford-fulkerson algorithm,
    # Edmonson-Karp and Dinic, but I didn't like them due to their naive usage
    # of DFS, which leads to a lot of node visits.
    #
    # I therefore choose to invert the capacities of the graph so that the
    # capacity any G[u][v] = c becomes 1/c in G_inverted.
    # This allows me to use the shortest path method to find the path with
    # most capacity in the first attempt, resulting in a significant reduction
    # of unassigned flow.
    #
    # By updating G_inverted, with the residual capacity, I can keep using the
    # shortest path, until the capacity is zero, whereby I remove the links
    # When the shortest path method returns 'No path' or when unassigned flow
    # is zero, I exit the algorithm.
    #
    # Even on small graphs, this method is very efficient, despite the overhead
    # of using shortest path. For very large graphs, this method outperforms
    # all other algorithms by orders of magnitude.
    # -----------------------------------------------------------------------

    edges = [(n1, n2, 1 / d) for n1, n2, d in graph.edges() if d > 0]
    inverted_graph = Graph(from_list=edges)  # create G_inverted.
    capacity_graph = Graph()  # Create structure to record capacity left.
    flow_graph = Graph()  # Create structure to record flows.

    while unassigned_flow:
        # 1. find the best path
        d, path = shortest_path(inverted_graph, start, end)
        if d == float('inf'):  # then there is no path, and we must exit.
            return total_flow, flow_graph
        # else: use the path and lookup the actual flow from the capacity graph.

        path_flow = min([min(d, capacity_graph.edge(s, e, default=float('inf')))
                         for s, e, d in graph.edges(path=path)])

        # 2. update the unassigned flow.
        unassigned_flow -= path_flow
        total_flow += path_flow

        # 3. record the flows and update the inverted graph, so that it is
        #    ready for the next iteration.
        edges = graph.edges(path)
        for n1, n2, d in edges:

            # 3.a. recording:
            v = flow_graph.edge(n1, n2, default=None)
            if v is None:
                flow_graph.add_edge(n1, n2, path_flow)
                c = graph.edge(n1, n2) - path_flow
            else:
                flow_graph.add_edge(n1, n2, value=v + path_flow)
                c = graph.edge(n1, n2) - (v + path_flow)
            capacity_graph.add_edge(n1, n2, c)

            # 3.b. updating:
            # if there is capacity left: update with new 1/capacity
            # else: remove node, as we can't do 1/zero.
            if c > 0:
                inverted_graph.add_edge(n1, n2, 1 / c)
            else:
                inverted_graph.del_edge(n1, n2)
    return total_flow, flow_graph


def maximum_flow_min_cut(graph, start, end):
    """
    Finds the edges in the maximum flow min cut.
    :param graph: Graph
    :param start: start
    :param end: end
    :return: list of edges
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError(f"{start} not in graph")
    if end not in graph:
        raise ValueError(f"{end} not in graph")

    flow, mfg = maximum_flow(graph, start, end)
    if flow == 0:
        return []

    working_graph = Graph(from_list=mfg.to_list())

    min_cut = []
    for n1 in mfg.breadth_first_walk(start, end):
        n2s = mfg.nodes(from_node=n1)
        for n2 in n2s:
            if graph.edge(n1, n2) - mfg.edge(n1, n2) == 0:
                working_graph.del_edge(n1, n2)
                min_cut.append((n1, n2))

    min_cut_nodes = set(working_graph.nodes(out_degree=0))
    min_cut_nodes.remove(end)
    min_cut = [(n1, n2) for (n1, n2) in min_cut if n1 in min_cut_nodes]
    return min_cut


def minimum_cost_flow_using_successive_shortest_path(costs, inventory, capacity=None):
    """
    Calculates the minimum cost flow solution using successive shortest path.
    :param costs: Graph with `cost per unit` as edge
    :param inventory: dict {node: stock, ...}
        stock < 0 is demand
        stock > 0 is supply
    :param capacity: None or Graph with `capacity` as edge.
        if capacity is None, capacity is assumed to be float('inf')
    :return: total costs, flow graph
    """
    if not isinstance(costs, Graph):
        raise TypeError(f"expected costs as Graph, not {type(costs)}")
    if not isinstance(inventory, dict):
        raise TypeError(f"expected inventory as dict, not {type(inventory)}")

    if not all(d >= 0 for s, e, d in costs.edges()):
        raise ValueError("The costs graph has negative edges. That won't work.")

    if not all(isinstance(v, (float, int)) for v in inventory.values()):
        raise TypeError("not all stock is numeric.")

    if capacity is None:
        capacity = Graph(from_list=[(s, e, float('inf')) for s, e, d in costs.edges()])
    else:
        if not isinstance(capacity, Graph):
            raise TypeError("Expected capacity as a Graph")
        if any(d < 0 for s, e, d in capacity.edges()):
            nn = [(s, e) for s, e, d in capacity.edges() if d < 0]
            raise ValueError(f"negative capacity on edges: {nn}")
        if {(s, e) for s, e, d in costs.edges()} != {(s, e) for s, e, d in capacity.edges()}:
            raise ValueError("cost and capacity have different links")

    # successive shortest path algorithm begins ...
    # ------------------------------------------------
    paths = costs.copy()  # initialise a copy of the cost graph so edges that
    # have exhausted capacities can be removed.
    flows = Graph()  # initialise F as copy with zero flow
    capacities = Graph()  # initialise C as a copy of capacity, so used capacity
    # can be removed.
    balance = [(v, k) for k, v in inventory.items() if v != 0]  # list with excess/demand, node id
    balance.sort()

    distances = paths.all_pairs_shortest_paths()

    while balance:  # while determine excess / imbalances:
        D, Dn = balance[0]  # pick node Dn where the demand D is greatest
        if D > 0:
            break  # only supplies left.
        balance = balance[1:]  # remove selection.

        supply_sites = [(distances[En][Dn], E, En) for E, En in balance if E > 0]
        if not supply_sites:
            break  # supply exhausted.
        supply_sites.sort()
        dist, E, En = supply_sites[0]  # pick nearest node En with excess E.
        balance.remove((E, En))  # maintain balance by removing the node.

        if E < 0:
            break  # no supplies left.
        if dist == float('inf'):
            raise Exception("bad logic: Case not checked for.")

        cost, path = paths.shortest_path(En, Dn)  # compute shortest path P from E to a node in demand D.

        # determine the capacity limit C on P:
        capacity_limit = min(capacities.edge(s, e, default=capacity.edge(s, e)) for s, e in zip(path[:-1], path[1:]))

        # determine L units to be transferred as min(demand @ D and the limit C)
        L = min(E, abs(D), capacity_limit)
        for s, e in zip(path[:-1], path[1:]):
            flows.add_edge(s, e, L + flows.edge(s, e, default=0))  # update F.
            new_capacity = capacities.edge(s, e, default=capacity.edge(s, e)) - L
            capacities.add_edge(s, e, new_capacity)  # update C

            if new_capacity == 0:  # remove the edge from potential solutions.
                paths.del_edge(s, e)
                distances = paths.all_pairs_shortest_paths()

        # maintain balance, in case there is excess or demand left.
        if E - L > 0:
            insort(balance, (E - L, En))
        if D + L < 0:
            insort(balance, (D + L, Dn))

    total_cost = sum(d * costs.edge(s, e) for s, e, d in flows.edges())
    return total_cost, flows


def tsp_branch_and_bound(graph):
    """
    Solve the traveling salesman's problem for the graph.
    :param graph: instance of class Graph
    :return: tour_length, path

    solution quality 100%
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    def lower_bound(graph, nodes):
        """ Calculates the lower bound of distances for given nodes. """
        L = []
        edges = set()
        for n in nodes:

            L2 = [(d, e) for s, e, d in graph.edges(from_node=n) if e in nodes - {n}]
            if not L2:
                continue
            L2.sort()

            for d, n2 in L2:
                if (n2, n) in edges:  # Solution is not valid as it creates a loop.
                    continue
                else:
                    edges.add((n, n2))  # remember!
                    L.append((n, n2, d))
                    break

        return L

    global_lower_bound = sum(d for n, n2, d in lower_bound(graph, set(graph.nodes())))

    q = []
    all_nodes = set(graph.nodes())

    # create initial tree.
    start = graph.nodes()[0]
    for start, end, distance in graph.edges(from_node=start):
        lb = lower_bound(graph, all_nodes - {start})
        dist = sum(d for s, e, d in lb)
        insort(q,
               (distance + dist,  # lower bound of distance.
                -2,  # number of nodes in tour.
                (start, end))  # locations visited.
               )

    hit, switch, q2 = 0, True, []
    while q:  # walk the tree.
        d, _, tour = q.pop(0)
        tour_set = set(tour)

        if tour_set == all_nodes:
            if hit < len(all_nodes):  # to overcome premature exit.
                hit += 1
                insort(q2, (d, tour))
                continue
            else:
                d, tour = q2.pop(0)
                assert d >= global_lower_bound, "Solution not possible."
                return d, list(tour[:-1])

        remaining_nodes = all_nodes - tour_set

        for n2 in remaining_nodes:

            new_tour = tour + (n2, )

            lb_set = remaining_nodes - {n2}
            if len(lb_set) > 1:
                lb_dists = lower_bound(graph, lb_set)
                lb = sum(d for n, n2, d in lb_dists)
                new_lb = graph.distance_from_path(new_tour) + lb
            elif len(lb_set) == 1:
                last_node = lb_set.pop()
                new_tour = new_tour + (last_node, tour[0])
                new_lb = graph.distance_from_path(new_tour)
            else:
                raise Exception("bad logic!")

            insort(q, (new_lb, -len(new_tour), new_tour))

    return float('inf'), []  # <-- exit path if not solvable.


def tsp_greedy(graph):
    """
    Solves the traveling salesman's problem for the graph.

    Runtime approximation: seconds = 10**(-5) * (nodes)**2.31
    Solution quality: Range 98.1% - 100% optimal.

    :param graph: instance of class Graph
    :return: tour_length, path
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    def shortest_links_first(graph):
        """ returns a list of (distance, node1, node2) with shortest on top."""
        c = combinations(graph.nodes(), 2)
        distances = [(graph.edge(a, b), a, b) for a, b in c if graph.edge(a, b)]
        distances.sort()
        return distances

    def join_endpoints(endpoints, a, b):
        """ Join segments [...,a] + [b,...] into one segment. Maintain `endpoints`.
        :param endpoints:
        :param a: node
        :param b: node
        :return:
        """
        a_seg, b_seg = endpoints[a], endpoints[b]
        if a_seg[-1] is not a:
            a_seg.reverse()
        if b_seg[0] is not b:
            b_seg.reverse()
        a_seg += b_seg
        del endpoints[a]
        del endpoints[b]
        endpoints[a_seg[0]] = endpoints[a_seg[-1]] = a_seg
        return a_seg

    def tsp_tour_length(graph, tour):
        """ The TSP tour length WITH return to the starting point."""
        return sum(graph.edge(tour[i - 1], tour[i]) for i in range(len(tour)))
        # note to above: If there's an error it's probably because the graph isn't
        # fully connected.

    def improve_tour(graph, tour):
        assert tour, "no tour to improve?"
        n = len(tour)

        # Return (i, j) pairs denoting tour[i:j] sub_segments of a tour of length N.
        sub_segments = [(i, i + length) for length in reversed(range(2, n)) for i in reversed(range(n - length + 1))]

        while True:
            improvements = {reverse_segment_if_improvement(graph, tour, i, j) for (i, j) in sub_segments}
            if improvements == {None} or len(improvements) == 0:
                return tour

    def reverse_segment_if_improvement(graph, tour, i, j):
        """If reversing tour[i:j] would make the tour shorter, then do it."""
        # Given tour [...A,B...C,D...], consider reversing B...C to get [...A,C...B,D...]
        a, b, c, d = tour[i - 1], tour[i], tour[j - 1], tour[j % len(tour)]
        # are old links (ab + cd) longer than new ones (ac + bd)? if so, reverse segment.
        A, B, C, D = graph.edge(a, b), graph.edge(c, d), graph.edge(a, c), graph.edge(b, d)
        # if all are not None and improvement is shorter than previous ...
        if all((A, B, C, D)) and A + B > C + D:
            tour[i:j] = reversed(tour[i:j])  # ..retain the solution.
            return True

    # The core TSP solver
    # -----------------------
    # 1. create a path using greedy algorithm (picks nearest peer)
    new_segment = []
    endpoints = {n: [n] for n in graph.nodes()}
    L = shortest_links_first(graph)
    for _, a, b in L:
        if a in endpoints and b in endpoints and endpoints[a] != endpoints[b]:
            new_segment = join_endpoints(endpoints, a, b)
            if len(new_segment) == len(graph.nodes()):
                break  # return new_segment
    assert len(new_segment) == len(graph.nodes()), "there's an unconnected component."
    first_tour = new_segment[:]
    first_path_length = tsp_tour_length(graph, first_tour)

    # 2. run improvement on the created path.
    improved_tour = improve_tour(graph, new_segment)
    assert set(graph.nodes()) == set(improved_tour)

    second_path_length = tsp_tour_length(graph, improved_tour)

    assert first_path_length >= second_path_length, "first path was better than improved tour?! {} {}".format(
        first_path_length, second_path_length
    )

    return second_path_length, improved_tour


def subgraph(graph, nodes):
    """ Creates a subgraph as a copy from the graph
    :param graph: class Graph
    :param nodes: set or list of nodes
    :return: new instance of Graph.
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if not isinstance(nodes, (set, list)):
        raise TypeError(f"expected nodes as a set or a list, not {type(nodes)}")

    node_set = set(nodes)
    G = object.__new__(graph.__class__)
    assert isinstance(G, BasicGraph)
    G.__init__()
    for n1 in nodes:
        obj = graph.node(n1)
        G.add_node(n1, obj)
        for n2 in graph.nodes(from_node=n1):
            if n2 in node_set:
                G.add_edge(n1, n2, graph.edge(n1, n2))
    return G


def is_subgraph(graph1, graph2):
    """
    Checks if graph1 is subgraph in graph2
    :param graph1: instance of Graph
    :param graph2: instance of Graph
    :return: boolean
    """
    if not isinstance(graph1, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph1)}")
    if not isinstance(graph2, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph2)}")

    if not set(graph1.nodes()).issubset(set(graph2.nodes())):
        return False
    if not set(graph1.edges()).issubset(set(graph2.edges())):
        return False
    return True


def is_partite(graph, n):
    """ Checks if graph is n-partite
    :param graph: class Graph
    :param n: int, number of partitions.
    :return: boolean and partitions as dict[colour] = set(nodes) or None.
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    if not isinstance(n, int):
        raise TypeError(f"Expected n as integer > 0, not {type(n)}")
    colours_and_nodes = {i: set() for i in range(n)}
    nodes_and_colours = {}
    n1 = set(graph.nodes()).pop()
    q = [n1]
    visited = set()
    colour = 0
    while q:
        n1 = q.pop()
        visited.add(n1)

        if n1 in nodes_and_colours:
            colour = nodes_and_colours[n1]
        else:
            colours_and_nodes[colour].add(n1)
            nodes_and_colours[n1] = colour

        next_colour = (colour + 1) % n
        neighbours = graph.nodes(from_node=n1) + graph.nodes(to_node=n1)
        for n2 in neighbours:
            if n2 in nodes_and_colours:
                if nodes_and_colours[n2] == colour:
                    return False, None
                # else:  pass  # it already has a colour and there is no conflict.
            else:  # if n2 not in nodes_and_colours:
                colours_and_nodes[next_colour].add(n2)
                nodes_and_colours[n2] = next_colour
                continue
            if n2 not in visited:
                q.append(n2)

    return True, colours_and_nodes


def has_cycles(graph):
    """ Checks if graph has a cycle
    :param graph: instance of class Graph.
    :return: bool
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    for n1, n2, d in graph.edges():
        if n1 == n2:
            return True
        if graph.is_connected(n2, n1):
            return True
    return False


def components(graph):
    """ Determines the components of the graph
    :param graph: instance of class Graph
    :return: list of sets of nodes. Each set is a component.
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    nodes = set(graph.nodes())
    sets_of_components = []
    while nodes:
        new_component = set()
        sets_of_components.append(new_component)
        n = nodes.pop()  # select random node
        new_component.add(n)  # add it to the new component.

        new_nodes = set(graph.nodes(from_node=n))
        new_nodes.update(set(graph.nodes(to_node=n)))
        while new_nodes:
            n = new_nodes.pop()
            new_component.add(n)
            new_nodes.update(set(n for n in graph.nodes(from_node=n) if n not in new_component))
            new_nodes.update(set(n for n in graph.nodes(to_node=n) if n not in new_component))
        nodes = nodes - new_component
    return sets_of_components


def network_size(graph, n1, degrees_of_separation=None):
    """ Determines the nodes within the range given by
    a degree of separation
    :param graph: Graph
    :param n1: start node
    :param degrees_of_separation: integer
    :return: set of nodes within given range
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if n1 not in graph:
        raise ValueError(f"{n1} not in graph")

    if degrees_of_separation is not None:
        if not isinstance(degrees_of_separation, int):
            raise TypeError(f"Expected degrees_of_separation to be integer, not {type(degrees_of_separation)}")

    network = {n1}
    q = set(graph.nodes(from_node=n1))

    scan_depth = 1
    while True:
        if not q:  # then there's no network.
            break

        if degrees_of_separation is not None:
            if scan_depth > degrees_of_separation:
                break

        new_q = set()
        for peer in q:
            if peer in network:
                continue
            else:
                network.add(peer)
                new_peers = set(graph.nodes(from_node=peer)) - network
                new_q.update(new_peers)
        q = new_q
        scan_depth += 1
    return network


def topological_sort(graph, key=None):
    """Return a generator of nodes in topologically sorted order.

    :param graph: Graph
    :param key: optional function for sortation.
    :return: Generator

    Topological sort (ordering) is a linear ordering of vertices.
    https://en.wikipedia.org/wiki/Topological_sorting

    Note: The algorithm does not check for loops before initiating
    the sortation, but raise Exception at the first conflict.
    This saves O(m+n) runtime.

    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    if key is None:
        def key(x): return x

    g2 = graph.copy()

    zero_in_degree = sorted(g2.nodes(in_degree=0), key=key)

    while zero_in_degree:
        for task in zero_in_degree:
            yield task  # <--- do something.

            g2.del_node(task)

        zero_in_degree = sorted(g2.nodes(in_degree=0), key=key)

    if g2.nodes():
        raise TypeError(f"Graph is not acyclic: Loop found: {g2.nodes()}")


def phase_lines(graph):
    """ Determines the phase lines of a directed graph.
    :param graph: Graph
    :return: dictionary with node id : phase in cut.

    Note: To transform the phaselines into a task sequence use
    the following recipe:

    tasks = defaultdict(set)
    for node, phase in phaselines(graph):
        tasks[phase].add(node)

    To obtain a sort stable tasks sequence use:

    for phase in sorted(tasks):
        print(phase, list(sorted(tasks[phase]))

    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    phases = {n: 0 for n in graph.nodes()}
    sinks = {n: set() for n in phases}  # sinks[e] = {s1,s2}
    edges = {n: set() for n in phases}
    for s, e, d in graph.edges():
        sinks[e].add(s)
        edges[s].add(e)

    cache = {}

    level = 0
    while sinks:
        sources = [e for e in sinks if not sinks[e]]  # these nodes have in_degree=0
        if not sources:
            raise AttributeError("The graph does not have any sinks.")
        for s in sources:
            phases[s] = level  # let's update the phase value
            del sinks[s]  # and let's remove their sink entry.
            # and remove their set item from the sinks dict
            for e in edges[s]:
                if e not in sinks:
                    continue
                sinks[e].discard(s)
                # but also check if their descendants are loops.
                for s2 in list(sinks[e]):

                    con = cache.get((e, s2))  # check if the edge has been seen before.
                    if con is None:
                        con = graph.is_connected(e, s2)  # if not seen before, search...
                        cache[(e, s2)] = con

                    if con:
                        sinks[e].discard(s2)
        level += 1

    return phases


def sources(graph, n):
    """ Determines the set of all upstream sources of node 'n' in a DAG.
    :param graph: Graph
    :param n: node for which the sources are sought.
    :return: set of nodes
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if n not in graph:
        raise ValueError(f"{n} not in graph")

    nodes = {n}
    q = deque([n])
    while q:
        new = q.popleft()
        for src in graph.nodes(to_node=new):
            if src not in nodes:
                nodes.add(src)
            if src not in q:
                q.append(src)
    nodes.remove(n)
    return nodes


def same_path(path1, path2):
    """ Compares two paths to verify whether they're the same despite being offset.
    Very useful when comparing results from TSP as solutions may be rotations of
    the same path.
    :param path1: list of nodes.
    :param path2: list of nodes.
    :return: boolean.
    """
    if not isinstance(path1, (list, set, tuple)):
        raise TypeError(f"Expected path1 as Iterable, not {type(path1)}")
    if not isinstance(path2, (list, set, tuple)):
        raise TypeError(f"Expected path2 as Iterable, not {type(path2)}")

    if path1 is path2:  # use id system to avoid work.
        return True
    if len(path1) != len(path2) or set(path1) != set(path2):
        return False

    starts = (ix for ix, n2 in enumerate(path2) if path1[0] == n2)
    return any(all(a == b for a, b in zip(path1, chain(path2[start:], path2[:start]))) for start in starts)


def adjacency_matrix(graph):
    """Converts directed graph to an adjacency matrix.
    :param graph:
    :return: dictionary

    The distance from a node to itself is 0 and distance from a node to
    an unconnected node is defined to be infinite. This does not mean that there
    is no path from a node to another via other nodes.

    Example:
        g = Graph(from_dict=
            {1: {2: 3, 3: 8, 5: -4},
             2: {4: 1, 5: 7},
             3: {2: 4},
             4: {1: 2, 3: -5},
             5: {4: 6}})

        adjacency_matrix(g)
        {1: {1: 0, 2: 3, 3: 8, 4: inf, 5: -4},
         2: {1: inf, 2: 0, 3: inf, 4: 1, 5: 7},
         3: {1: inf, 2: 4, 3: 0, 4: inf, 5: inf},
         4: {1: 2, 2: inf, 3: -5, 4: 0, 5: inf},
         5: {1: inf, 2: inf, 3: inf, 4: 6, 5: 0}}
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    return {v1: {v2: 0 if v1 == v2 else graph.edge(v1, v2, default=float('inf'))
                 for v2 in graph.nodes()}
            for v1 in graph.nodes()}


def all_pairs_shortest_paths(graph):
    """Find the cost of the shortest path between every pair of vertices in a
    weighted graph. Uses the Floyd-Warshall algorithm.

    Example:
        inf = float('inf')
        g = Graph(from_dict=(
            {0: {0: 0,   1: 1,   2: 4},
             1: {0: inf, 1: 0,   2: 2},
             2: {0: inf, 1: inf, 2: 0}})

        fw(g)
        {0: {0: 0,   1: 1,   2: 3},
        1: {0: inf, 1: 0,   2: 2},
        2: {0: inf, 1: inf, 2: 0}}

        h = {1: {2: 3, 3: 8, 5: -4},
             2: {4: 1, 5: 7},
             3: {2: 4},
             4: {1: 2, 3: -5},
             5: {4: 6}}

        fw(adj(h)) #
            {1: {1: 0, 2:  1, 3: -3, 4: 2, 5: -4},
             2: {1: 3, 2:  0, 3: -4, 4: 1, 5: -1},
             3: {1: 7, 2:  4, 3:  0, 4: 5, 5:  3},
             4: {1: 2, 2: -1, 3: -5, 4: 0, 5: -2},
             5: {1: 8, 2:  5, 3:  1, 4: 6, 5:  0}}
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    g = graph.adjacency_matrix()
    assert isinstance(g, dict), "previous function should have returned a dict."
    vertices = g.keys()

    for v2 in vertices:
        g = {v1: {v3: min(g[v1][v3], g[v1][v2] + g[v2][v3])
                  for v3 in vertices}
             for v1 in vertices}
    return g


def minsum(graph):
    """ finds the mode(s) that have the smallest sum of distance to all other nodes. """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    adj_mat = graph.all_pairs_shortest_paths()
    for n in adj_mat:
        adj_mat[n] = sum(adj_mat[n].values())
    smallest = min(adj_mat.values())
    return [k for k, v in adj_mat.items() if v == smallest]


def minmax(graph):
    """ finds the node(s) with shortest distance to all other nodes. """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    adj_mat = graph.all_pairs_shortest_paths()
    for n in adj_mat:
        adj_mat[n] = max(adj_mat[n].values())
    smallest = min(adj_mat.values())
    return [k for k, v in adj_mat.items() if v == smallest]


def shortest_tree_all_pairs(graph):
    """
       'minimize the longest distance between any pair'

    Note: This algorithm is not shortest path as it jumps
    to a new branch when it has exhausted a branch in the tree.
    :return: path
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    g = all_pairs_shortest_paths(graph)
    assert isinstance(g, dict)

    distance = float('inf')
    best_starting_point = -1
    # create shortest path gantt diagram.
    for start_node in g.keys():
        if start_node in g:
            dist = sum(v for k, v in g[start_node].items())
            if dist < distance:
                best_starting_point = start_node
            # else: skip the node as it's isolated.
    g2 = g[best_starting_point]  # {1: 0, 2: 1, 3: 2, 4: 3}

    inv_g2 = {}
    for k, v in g2.items():
        if v not in inv_g2:
            inv_g2[v] = set()
        inv_g2[v].add(k)

    all_nodes = set(g.keys())
    del g
    path = []
    while all_nodes and inv_g2.keys():
        v_nearest = min(inv_g2.keys())
        for v in inv_g2[v_nearest]:
            all_nodes.remove(v)
            path.append(v)
        del inv_g2[v_nearest]
    return path


def has_path(graph, path):
    """ checks if path exists is graph
    :param graph: instance of Graph
    :param path: list of nodes
    :return: boolean
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if not isinstance(path, (list, tuple)):
        raise TypeError(f"Expected list or tuple, not {type(path)}")
    v1 = path[0]
    for v2 in path[1:]:
        if graph.edge(v1, v2) is None:
            return False
        else:
            v1 = v2
    return True


def all_simple_paths(graph, start, end):
    """
    finds all simple (non-looping) paths from start to end
    :param start: node
    :param end: node
    :return: list of paths
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError("start not in graph.")
    if end not in graph:
        raise ValueError("end not in graph.")
    if start == end:
        raise ValueError("start is end")

    if not graph.is_connected(start, end):
        return []

    paths = []
    q = deque([(start,)])
    while q:
        path = q.popleft()
        for s, e, d in graph.edges(from_node=path[0]):
            if e in path:
                continue
            new_path = (e,) + path
            if e == end:
                paths.append(new_path)
            else:
                q.append(new_path)
    return [list(reversed(p)) for p in paths]


def all_paths(graph, start, end):
    """ finds all paths from start to end by traversing each fork once only.
    :param graph: instance of Graph
    :param start: node
    :param end: node
    :return: list of paths unique from start to end.
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError("start not in graph.")
    if end not in graph:
        raise ValueError("end not in graph.")
    if start == end:
        raise ValueError("start is end")

    cache = {}
    if not graph.is_connected(start, end):
        return []
    paths = [(start,)]
    q = deque([start])
    skip_list = set()
    while q:
        n1 = q.popleft()
        if n1 == end:
            continue

        n2s = graph.nodes(from_node=n1)
        new_paths = [p for p in paths if p[-1] == n1]
        for n2 in n2s:
            if n2 in skip_list:
                continue
            n3s = graph.nodes(from_node=n2)

            con = cache.get((n2, n1))
            if con is None:
                con = graph.is_connected(n2, n1)
                cache[(n2, n1)] = con

            if len(n3s) > 1 and con:
                # it's a fork and it's a part of a loop!
                # is the sequence n2,n3 already in the path?
                for n3 in n3s:
                    for path in new_paths:
                        a = [n2, n3]
                        if any(all(path[i+j] == a[j] for j in range(len(a))) for i in range(len(path))):
                            skip_list.add(n3)

            for path in new_paths:
                if path in paths:
                    paths.remove(path)

                new_path = path + (n2,)
                if new_path not in paths:
                    paths.append(new_path)

            if n2 not in q:
                q.append(n2)

    paths = [list(p) for p in paths if p[-1] == end]
    return paths


def degree_of_separation(graph, n1, n2):
    """ Calculates the degree of separation between 2 nodes."""
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if n1 not in graph:
        raise ValueError("n1 not in graph.")
    if n2 not in graph:
        raise ValueError("n2 not in graph.")

    assert n1 in graph.nodes()
    p = breadth_first_search(graph, n1, n2)
    return len(p)-1


def loop(graph, start, mid, end=None):
    """ Returns a loop passing through a defined mid-point and returning via a different set of nodes to the outward
        journey. If end is None we return to the start position. """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError("start not in graph.")
    if mid not in graph:
        raise ValueError("mid not in graph.")
    if end is not None and end not in graph:
        raise ValueError("end not in graph.")

    _, p = graph.shortest_path(start, mid)
    g2 = graph.copy()
    if end is not None:
        for n in p[:-1]:
            g2.del_node(n)
        _, p2 = g2.shortest_path(mid, end)
    else:
        for n in p[1:-1]:
            g2.del_node(n)
        _, p2 = g2.shortest_path(mid, start)
    lp = p + p2[1:]
    return lp


class ScanThread(object):
    __slots__ = ['cost', 'n1', 'path']

    """ search thread for bidirectional search """
    def __init__(self, cost, n1, path=()):
        if not isinstance(path, tuple):
            raise TypeError(f"Expected a tuple, not {type(path)}")
        self.cost = cost
        self.n1 = n1
        self.path = path

    def __lt__(self, other):
        return self.cost < other.cost

    def __str__(self):
        return f"{self.cost}:{self.path}"


class SPLength(object):
    def __init__(self):
        self.value = float('inf')


class BiDirectionalSearch(object):
    """ data structure for organizing bidirectional search """
    forward = True
    backward = False

    def __str__(self):
        if self.forward == self.direction:
            return "forward scan"
        return "backward scan"

    def __init__(self, graph, start, direction=True, avoids=None):
        """
        :param graph: class Graph.
        :param start: first node in the search.
        :param direction: bool
        :param avoids: nodes that cannot be a part of the solution.
        """
        if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
            raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
        if start not in graph:
            raise ValueError("start not in graph.")
        if not isinstance(direction, bool):
            raise TypeError(f"Expected boolean, not {type(direction)}")
        if avoids is None:
            self.avoids = frozenset()
        elif not isinstance(avoids, (frozenset, set, list)):
            raise TypeError(f"Expect obstacles as set or frozenset, not {type(avoids)}")
        else:
            self.avoids = frozenset(avoids)

        self.q = []
        self.q.append(ScanThread(cost=0, n1=start))
        self.graph = graph
        self.boundary = set()  # visited.
        self.mins = {start: 0}
        self.paths = {start: ()}
        self.direction = direction
        self.sp = ()
        self.sp_length = float('inf')

    def update(self, sp, sp_length):
        if sp_length > self.sp_length:
            raise ValueError("Bad logic!")
        self.sp = sp
        self.sp_length = sp_length

    def search(self, other):
        assert isinstance(other, BiDirectionalSearch)
        if not self.q:
            return

        sp, sp_length = self.sp, self.sp_length

        st = self.q.pop(0)
        assert isinstance(st, ScanThread)
        if st.cost > self.sp_length:
            return

        self.boundary.add(st.n1)

        if st.n1 in other.boundary:  # if there's an intercept between the two searches ...
            if st.cost + other.mins[st.n1] < self.sp_length:
                sp_length = st.cost + other.mins[st.n1]
                if self.direction == self.forward:
                    sp = tuple(reversed(st.path)) + (st.n1,) + other.paths[st.n1]
                else:  # direction == backward:
                    sp = tuple(reversed(other.paths[st.n1])) + (st.n1,) + st.path

                self.q = [a for a in self.q if a.cost < sp_length]

        if self.direction == self.forward:
            edges = sorted((d, e) for s, e, d in self.graph.edges(from_node=st.n1) if e not in self.avoids)
        else:
            edges = sorted((d, s) for s, e, d in self.graph.edges(to_node=st.n1) if s not in self.avoids)

        for dist, n2 in edges:
            n2_dist = st.cost + dist
            if n2_dist > self.sp_length:  # no point pursuing as the solution is worse.
                continue
            if n2 in other.mins and n2_dist + other.mins[n2] > self.sp_length:  # already longer than lower bound.
                continue

            # at this point we can't dismiss that n2 will lead to a better solution, so we retain it.
            prev = self.mins.get(n2, None)
            if prev is None or n2_dist < prev:
                self.mins[n2] = n2_dist
                path = (st.n1,) + st.path
                self.paths[n2] = path
                insort(self.q, ScanThread(n2_dist, n2, path))

        self.update(sp, sp_length)
        other.update(sp, sp_length)


def shortest_path_bidirectional(graph, start, end, avoids=None):
    """ Bidirectional search using lower bound.
    :param graph: Graph
    :param start: start node
    :param end: end node
    :param avoids: nodes that cannot be a part of the shortest path.
    :return: shortest path

    In Section 3.4.6 of Artificial Intelligence: A Modern Approach, Russel and
    Norvig write:

    Bidirectional search is implemented by replacing the goal test with a check
    to see whether the frontiers of the two searches intersect; if they do,
    a solution has been found. It is important to realize that the first solution
    found may not be optimal, even if the two searches are both breadth-first;
    some additional search is required to make sure there isn't a shortcut
    across the gap.

    To overcome this limit for weighted graphs, I've added a lower bound, so
    that when the two searches intersect, the lower bound is updated and the
    lower bound path is stored. In subsequent searches any path shorter than
    the lower bound, leads to an update of the lower bound and shortest path.
    The algorithm stops when all nodes on the frontier exceed the lower bound.

    ----------------

    The algorithms works as follows:

    Lower bound = float('infinite')
    shortest path = None

    Two queues (forward scan and backward scan) are initiated with respectively
    the start and end node as starting point for each scan.

    while there are nodes in the forward- and backward-scan queues:
        1. select direction from (forward, backward) in alternations.
        2. pop the top item from the queue of the direction.
           (The top item contains the node N that is nearest the starting point for the scan)
        3. Add the node N to the scan-directions frontier.
        4. If the node N is in the other directions frontier:
            the path distance from the directions _start_ to the _end_ via
            the point of intersection (N), is compared with the lower bound.

            If the path distance is less than the lower bound:
            *lower bound* is updated with path distance, and,
            the *shortest path* is recorded.

        5. for each node N2 (to N if backward, from N if forward):
            the distance D is accumulated

            if D > lower bound, the node N2 is ignored.

            if N2 is within the other directions frontier and D + D.other > lower bound, the node N2 is ignored.

            otherwise:
                the path P recorded
                and N2, D and P are added to the directions scan queue

    The algorithm terminates when the scan queues are exhausted.
    ----------------

    Given that:
    - `c` is the connectivity of the nodes in the graph,
    - `R` is the length of the path,
    the explored solution landscape can be estimated as:

    A = c * (R**2), for single source shortest path
    A = c * 2 * (1/2 * R) **2, for bidirectional shortest path
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")
    if start not in graph:
        raise ValueError("start not in graph.")
    if end not in graph:
        raise ValueError("end not in graph.")

    forward = BiDirectionalSearch(graph, start=start, direction=BiDirectionalSearch.forward, avoids=avoids)
    backward = BiDirectionalSearch(graph, start=end, direction=BiDirectionalSearch.backward, avoids=avoids)

    while any((forward.q, backward.q)):
        forward.search(other=backward)
        backward.search(other=forward)

    return forward.sp_length, list(forward.sp)


class ShortestPathCache(object):
    """
    Data structure optimised for repeated calls to shortest path.
    Used by shortest path when using keyword `memoize=True`
    """
    def __init__(self, graph):
        if not isinstance(graph, Graph):
            raise TypeError(f"Expected type Graph, not {type(graph)}")
        self.graph = graph
        self.cache = {}
        self.repeated_cache = {}

    def _update_cache(self, path):
        """ private method for updating the cache for future lookups.
        :param path: tuple of nodes

        Given a shortest path, all steps along the shortest path,
        also constitute the shortest path between each pair of steps.
        """
        if not isinstance(path, (list, tuple)):
            raise TypeError
        b = len(path)
        if b < 2:
            return
        if b == 2:
            dist = self.graph.distance_from_path(path)
            self.cache[(path[0], path[-1])] = (dist, tuple(path))

        for a, _ in enumerate(path):
            section = tuple(path[a:b - a])
            if len(section) < 3:
                break
            dist = self.graph.distance_from_path(section)
            self.cache[(section[0], section[-1])] = (dist, section)

        for ix, start in enumerate(path[1:-1]):
            section = tuple(path[ix:])
            dist = self.graph.distance_from_path(section)
            self.cache[(section[0], section[-1])] = (dist, section)

    def shortest_path(self, start, end, avoids=None):
        """ Shortest path method that utilizes caching and bidirectional search """
        if start not in self.graph:
            raise ValueError("start not in graph.")
        if end not in self.graph:
            raise ValueError("end not in graph.")
        if avoids is None:
            pass
        elif not isinstance(avoids, (frozenset, set, list)):
            raise TypeError(f"Expect obstacles as None, set or frozenset, not {type(avoids)}")
        else:
            avoids = frozenset(avoids)

        d = 0 if start == end else None
        p = ()

        if isinstance(avoids, frozenset):
            # as avoids can be volatile, it is not possible to benefit from the caching
            # methodology. This does however not mean that we need to forfeit the benefit
            # of bidirectional search.
            hash_key = hash(avoids)
            d, p = self.repeated_cache.get((start, end, hash_key), (None, None))
            if d is None:
                d, p = shortest_path_bidirectional(self.graph, start, end, avoids=avoids)
                self.repeated_cache[(start, end, hash_key)] = (d, p)
        else:
            if d is None:  # is it cached?
                d, p = self.cache.get((start, end), (None, None))

            if d is None:  # search for it.
                _, p = shortest_path_bidirectional(self.graph, start, end)
                self._update_cache(p)
                d, p = self.cache[(start, end)]

        return d, list(p)


class Graph(BasicGraph):
    """
    Graph is the base graph that all methods use.

    For methods, please see the documentation on the
    individual functions, by importing them separately.

    """

    def __init__(self, from_dict=None, from_list=None):
        super().__init__(from_dict=from_dict, from_list=from_list)
        self._cache = None

    def shortest_path(self, start, end, memoize=False, avoids=None):
        """
        :param start: start node
        :param end: end node
        :param memoize: boolean (stores paths in a cache for faster repeated lookup)
        :param avoids: optional. A frozen set of nodes that cannot be on the path.
        :return: distance, path as list
        """
        if not memoize:
            return shortest_path(graph=self, start=start, end=end, avoids=avoids)

        if self._cache is None:
            self._cache = ShortestPathCache(graph=self)
        return self._cache.shortest_path(start, end, avoids=avoids)

    def shortest_path_bidirectional(self, start, end):
        """
        :param start: start node
        :param end: end node
        :return: distance, path as list
        """
        return shortest_path_bidirectional(self, start, end)

    def breadth_first_search(self, start, end):
        """ Determines the path with fewest nodes.
        :param start: start node
        :param end: end nodes
        :return: nodes, path as list
        """
        return breadth_first_search(graph=self, start=start, end=end)

    def breadth_first_walk(self, start, end=None, reversed_walk=False):
        """
        :param start: start node
        :param end: end node
        :param reversed_walk: if True, the BFS walk is backwards.
        :return: generator for breadth-first walk
        """
        return breadth_first_walk(graph=self, start=start, end=end, reversed_walk=reversed_walk)

    def distance_map(self, starts=None, ends=None, reverse=False):
        """ Maps the shortest path distance from any start to any end.
        :param graph: instance of Graph
        :param starts: node or (set,list,tuple) of nodes
        :param ends: None (exhaustive map), node or (set,list,tuple) of nodes
        :param reverse: boolean, if True follows edges backwards.
        :return: dictionary with {node: distance from start}
        """
        return distance_map(self, starts, ends, reverse)

    def depth_first_search(self, start, end):
        """
        Finds a path from start to end using DFS.
        :param start: start node
        :param end: end node
        :return: path
        """
        return depth_first_search(graph=self, start=start, end=end)

    def depth_scan(self, start, criteria):
        """
        traverses the descendants of node `start` using callable `criteria` to determine
        whether to terminate search along each branch in `graph`.

        :param start: start node
        :param criteria: function to terminate scan along a branch must return bool
        :return: set of nodes
        """
        return depth_scan(graph=self, start=start, criteria=criteria)

    def distance_from_path(self, path):
        """
        :param path: list of nodes
        :return: distance along the path.
        """
        return distance(graph=self, path=path)

    def maximum_flow(self, start, end):
        """ Determines the maximum flow of the graph between
        start and end.
        :param start: node (source)
        :param end: node (sink)
        :return: flow, graph of flow.
        """
        return maximum_flow(self, start, end)

    def maximum_flow_min_cut(self, start, end):
        """
        Finds the edges in the maximum flow min cut.
        :param start: start
        :param end: end
        :return: list of edges
        """
        return maximum_flow_min_cut(self, start, end)

    def minimum_cost_flow(self, inventory, capacity=None):
        """
        :param self: Graph with `cost per unit` as edge
        :param inventory: dict {node: stock, ...}
            stock < 0 is demand
            stock > 0 is supply
        :param capacity: None or Graph with `capacity` as edge.
        :return: total costs, graph of flows in solution.
        """
        return minimum_cost_flow_using_successive_shortest_path(self, inventory, capacity)

    def solve_tsp(self, method='greedy'):
        """ solves the traveling salesman problem for the graph
        (finds the shortest path through all nodes)

        :param method: str: 'greedy'

        options:
            'greedy' see tsp_greedy
            'bnb' see tsp_branch_and_bound

        :return: tour length (path+return to starting point),
                 path travelled.
        """
        methods = {
            'greedy': tsp_greedy,
            'bnb': tsp_branch_and_bound
        }
        solver = methods.get(method, 'greedy')
        return solver(self)

    def subgraph_from_nodes(self, nodes):
        """
        constructs a copy of the graph containing only the
        listed nodes (and their links)
        :param nodes: list of nodes
        :return: class Graph
        """
        return subgraph(graph=self, nodes=nodes)

    def is_subgraph(self, other):
        """ Checks if self is a subgraph in other.
        :param other: instance of Graph
        :return: boolean
        """
        return is_subgraph(self, other)

    def is_partite(self, n=2):
        """ Checks if self is n-partite
        :param n: int the number of partitions.
        :return: tuple: boolean, partitions as dict
                        (or None if graph isn't n-partite)
        """
        return is_partite(self, n)

    def has_cycles(self):
        """ Checks if the graph has a cycle
        :return: bool
        """
        return has_cycles(graph=self)

    def components(self):
        """ Determines the number of components
        :return: list of sets of nodes. Each set is a component.
        """
        return components(graph=self)

    def network_size(self, n1, degrees_of_separation=None):
        """ Determines the nodes within the range given by
        a degree of separation
        :param n1: start node
        :param degrees_of_separation: integer
        :return: set of nodes within given range
        """
        return network_size(self, n1, degrees_of_separation)

    def phase_lines(self):
        """ Determines the phase lines (cuts) of the graph
        :returns: dictionary with phase: nodes in phase
        """
        return phase_lines(self)

    def sources(self, n):
        """ Determines the DAG sources of node n """
        return sources(graph=self, n=n)

    def topological_sort(self, key=None):
        """ Returns a generator for the topological order"""
        return topological_sort(self, key=key)

    def critical_path(self):
        f"""{critical_path.__doc__}"""
        return critical_path(self)

    def critical_path_minimize_for_slack(self):
        f"""{critical_path_minimize_for_slack.__doc__}"""
        return critical_path_minimize_for_slack(self)

    @staticmethod
    def same_path(p1, p2):
        """ compares two paths to determine if they're the same, despite
        being in different order.

        :param p1: list of nodes
        :param p2: list of nodes
        :return: boolean
        """
        return same_path(p1, p2)

    def adjacency_matrix(self):
        """
        Converts directed graph to an adjacency matrix.
        Note: The distance from a node to itself is 0 and distance from a node to
        an unconnected node is defined to be infinite. This does not mean that there
        is no path from a node to another via other nodes.
        :return: dict
        """
        return adjacency_matrix(graph=self)

    def minsum(self):
        """ Finds the mode(s) that have the smallest sum of distance to all other nodes.
        :return: list of nodes
        """
        return minsum(self)

    def minmax(self):
        """ Finds the node(s) with shortest distance to all other nodes.
        :return: list of nodes
        """
        return minmax(self)

    def all_pairs_shortest_paths(self):
        """
        Find the cost of the shortest path between every pair of vertices in a
        weighted graph. Uses the Floyd-Warshall algorithm.
        :return: dict {node 1: {node 2: distance}, ...}
        """
        return all_pairs_shortest_paths(graph=self)

    def shortest_tree_all_pairs(self):
        """
        :return:
        """
        return shortest_tree_all_pairs(graph=self)

    def has_path(self, path):
        """
        :param path: list of nodes
        :return: boolean, if the path is in G.
        """
        return has_path(graph=self, path=path)

    def all_simple_paths(self, start, end):
        """
        finds all simple (non-looping) paths from start to end
        :param start: node
        :param end: node
        :return: list of paths
        """
        return all_simple_paths(self, start, end)

    def all_paths(self, start, end):
        """ finds all paths from start to end by traversing each fork once only.
        :param start: node
        :param end: node
        :return: list of paths
        """
        return all_paths(graph=self, start=start, end=end)

    def degree_of_separation(self, n1, n2):
        """ determines the degree of separation between 2 nodes
        :param n1: node
        :param n2: node
        :return: degree
        """
        return degree_of_separation(self, n1, n2)

    def loop(self, start, mid, end=None):
        """ finds a looped path via a mid-point
        :param start: node
        :param mid: node, midpoint for loop.
        :param end: node
        :return: path as list
        """
        return loop(self, start, mid, end)


class Graph3D(Graph):
    """ a graph where all (x,y)-positions are unique. """

    def __init__(self, from_dict=None, from_list=None):
        super().__init__(from_dict=from_dict, from_list=from_list)

    def copy(self):
        g = Graph3D(from_dict=self.to_dict())
        return g

    # spatial only function
    # ---------------------
    @staticmethod
    def _check_tuples(n1):
        if not isinstance(n1, tuple):
            raise TypeError(f"expected tuple, not {type(n1)}")
        if len(n1) != 3:
            raise ValueError(f"expected tuple in the form as (x,y,z), got {n1}")
        if not all(isinstance(i, (float, int)) for i in n1):
            raise TypeError(f"expected all values to be integer or float, but got {n1}")

    @staticmethod
    def distance(n1, n2):
        """ returns the distance between to xyz tuples coordinates
        :param n1: (x,y,z)
        :param n2: (x,y,z)
        :return: float
        """
        Graph3D._check_tuples(n1)
        Graph3D._check_tuples(n2)
        (x1, y1, z1), (x2, y2, z2) = n1, n2
        a = abs(x2 - x1)
        b = abs(y2 - y1)
        c = abs(z2 - z1)
        return (a * a + b * b + c * c) ** (1 / 2)

    def add_edge(self, n1, n2, value=None, bidirectional=False):
        self._check_tuples(n1)
        self._check_tuples(n2)
        assert value is not None
        super().add_edge(n1, n2, value, bidirectional)

    def add_node(self, node_id, obj=None):
        self._check_tuples(node_id)
        super().add_node(node_id, obj)
        """
        :param node_id: any hashable node.
        :param obj: any object that the node should refer to.

        PRO TIP: To retrieve the node obj use g.node(node_id)
        """
        self._nodes[node_id] = obj

    def n_nearest_neighbours(self, node_id, n=1):
        """ returns the node id of the `n` nearest neighbours. """
        self._check_tuples(node_id)
        if not isinstance(n, int):
            raise TypeError(f"expected n to be integer, not {type(n)}")
        if n < 1:
            raise ValueError(f"expected n >= 1, not {n}")

        d = [(self.distance(n1=node_id, n2=n), n) for n in self.nodes() if n != node_id]
        d.sort()
        if d:
            return [b for a, b in d][:n]
        return None

    def plot(self, nodes=True, edges=True, rotation='xyz', maintain_aspect_ratio=False):
        """ plots nodes and links using matplotlib3
        :param nodes: bool: plots nodes
        :param edges: bool: plots edges
        :param rotation: str: set view point as one of [xyz,xzy,yxz,yzx,zxy,zyx]
        :param maintain_aspect_ratio: bool: rescales the chart to maintain aspect ratio.
        :return: None. Plots figure.
        """
        return plot_3d(self, nodes, edges, rotation, maintain_aspect_ratio)


class Task(object):
    """ Helper for critical path method """
    __slots__ = ['task_id', 'duration',
                 'earliest_start', 'earliest_finish',
                 'latest_start', 'latest_finish']

    def __init__(self,
                 task_id, duration,
                 earliest_start=0, latest_start=0,
                 earliest_finish=float('inf'), latest_finish=float('inf')):
        self.task_id = task_id
        self.duration = duration
        self.earliest_start = earliest_start
        self.latest_start = latest_start
        self.earliest_finish = earliest_finish
        self.latest_finish = latest_finish

    def __eq__(self, other):
        if not isinstance(other, Task):
            raise TypeError(f"can't compare {type(other)} with {type(self)}")
        if any((
                self.task_id != other.task_id,
                self.duration != other.duration,
                self.earliest_start != other.earliest_start,
                self.latest_start != other.latest_start,
                self.earliest_finish != other.earliest_finish,
                self.latest_finish != other.latest_finish
        )):
            return False
        return True

    @property
    def slack(self):
        return self.latest_finish - self.earliest_finish

    @property
    def args(self):
        return (self.task_id, self.duration,
                self.earliest_start, self.latest_start,
                self.earliest_finish, self.latest_finish)

    def __repr__(self):
        return f"{self.__class__.__name__}{self.args}"

    def __str__(self):
        return self.__repr__()


def critical_path(graph):
    """
    The critical path method determines the
    schedule of a set of project activities that results in
    the shortest overall path.

    :param graph: acyclic graph where:
        nodes are task_id and node_obj is a value
        edges determines dependencies
        (see example below)

    :return: critical path length, schedule.
        schedule is list of Tasks

    Recipes:
    (1) Setting up the graph:

        tasks = {'A': 10, 'B': 20, 'C': 5, 'D': 10, 'E': 20, 'F': 15, 'G': 5, 'H': 15}
        dependencies = [
            ('A', 'B'),
            ('B', 'C'),
            ('C', 'D'),
            ('D', 'E'),
            ('A', 'F'),
            ('F', 'G'),
            ('G', 'E'),
            ('A', 'H'),
            ('H', 'E'),
        ]

        g = Graph()
        for task, duration in tasks.items():
            g.add_node(task, obj=duration)
        for n1, n2 in dependencies:
            g.add_edge(n1, n2, 0)
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    # 1. A topologically sorted list of nodes is prepared (topo.order).
    order = list(topological_sort(graph))  # this will raise if there are loops.

    d = {}
    critical_path_length = 0

    for task_id in order:  # 2. forward pass:
        predecessors = [d[t] for t in graph.nodes(to_node=task_id)]
        duration = graph.node(task_id)
        if not isinstance(duration, (float,int)):
            raise ValueError(f"Expected task {task_id} to have a numeric duration, but got {type(duration)}")
        t = Task(task_id=task_id, duration=duration)
        # 1. the earliest start and earliest finish is determined in topological order.
        t.earliest_start = max(t.earliest_finish for t in predecessors) if predecessors else 0
        t.earliest_finish = t.earliest_start + t.duration
        d[task_id] = t
        # 2. the path length is recorded.
        critical_path_length = max(t.earliest_finish, critical_path_length)

    for task_id in reversed(order):  # 3. backward pass:
        successors = [d[t] for t in graph.nodes(from_node=task_id)]
        t = d[task_id]
        # 1. the latest start and finish is determined in reverse topological order
        t.latest_finish = min(t.latest_start for t in successors) if successors else critical_path_length
        t.latest_start = t.latest_finish - t.duration

    return critical_path_length, d


def critical_path_minimize_for_slack(graph):
    """
    Determines the critical path schedule and attempts to minimise
    the concurrent resource requirements by inserting the minimal
    number of artificial dependencies.
    """
    if not isinstance(graph, (BasicGraph, Graph, Graph3D)):
        raise TypeError(f"Expected BasicGraph, Graph or Graph3D, not {type(graph)}")

    cpl, schedule = critical_path(graph)
    phases = phase_lines(graph)

    new_graph = graph.copy()
    slack_nodes = [t for k, t in schedule.items() if t.slack > 0]  # slack == 0 is on the critical path.
    slack_nodes.sort(reverse=True, key=lambda t: t.slack)  # most slack on top as it has more freedom to move.
    slack_node_ids = [t.task_id for t in slack_nodes]
    slack = sum(t.slack for t in schedule.values())

    # Part 1. Use a greedy algorithm to determine an initial solution.
    new_edges = []
    stop = False
    for node in sorted(slack_nodes, reverse=True, key=lambda t: t.duration):
        if stop:
            break
        # identify any option for insertion:
        n1 = node.task_id
        for n2 in (n2 for n2 in slack_node_ids if  # ...not on critical path
                   n2 != n1 and  # ...not pointing to the source
                   phases[n2] >= phases[n1] and  # ...downstream
                   new_graph.edge(n2, n1) is None and  # ... not creating a cycle.
                   new_graph.edge(n1, n2) is None):  # ... not already a dependency.

            new_graph.add_edge(n1, n2)
            new_edges.append((n1, n2))

            cpl2, schedule2 = critical_path(new_graph)
            if cpl2 != cpl:  # the critical path is not allowed to be longer. Abort.
                new_graph.del_edge(n1, n2)
                new_edges.remove((n1, n2))
                continue

            slack2 = sum(t.slack for t in schedule2.values())

            if slack2 < slack and cpl == cpl2:  # retain solution!
                slack = slack2
                if slack == 0:  # an optimal solution has been found!
                    stop = True
                    break
            else:
                new_graph.del_edge(n1, n2)
                new_edges.remove((n1, n2))

    # Part 2. Purge non-effective edges.
    for edge in new_graph.edges():
        n1, n2, _ = edge
        if graph.edge(n1, n2) is not None:
            continue  # it's an original edge.
        # else: it's an artificial edge:
        new_graph.del_edge(n1, n2)
        cpl2, schedule2 = critical_path(new_graph)
        slack2 = sum(t.slack for t in schedule2.values())

        if cpl2 == cpl and slack2 == slack:
            continue  # the removed edge had no effect and just made the graph more complicated.
        else:
            new_graph.add_edge(n1, n2)

    return new_graph

